/*******************************************************************************

        @file Format.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        
        @version        Initial version; April 2005

        @author         Kris


*******************************************************************************/

module mango.format.Format;

private import  std.stdarg;

private import  mango.sys.Type;

private import  mango.format.Int,
                mango.format.Long,
                mango.format.Styled;

/******************************************************************************

        Utility functions for styled, readable, output.

        While these functions are all static, they are encapsulated within 
        a class inheritance to preserve some namespace cohesion. One might 
        use structs for encapsualtion instead, but then inheritance would 
        be lost. Note that the root class, Styled, is abstract to prevent 
        accidental instantiation of these classes.

        See Styled.parse() for the list of format specifiers.

******************************************************************************/

class Format : Styled
{       
        /**********************************************************************

                Default styles for supported types. This is used to 
                format discrete values, or those provided without a 
                specified format-string. 

        **********************************************************************/

        static ubyte DefaultStyle[] = [
                                      's', 'd', 'u', 'd', 'u', 
                                      'd', 'u', 'd', 'u', 'f', 
                                      'f', 'f', 's', 's', 's', 
                                      'x'
                                      ];


        /**********************************************************************

                Equivalent to vsprintf(). State is maintained on the
                stack only, making this thread-safe. You may supply
                a workspace buffer as an optional initialization arg,
                which should typically also be allocated on the stack.

                See Styled.parse() for the list of format specifiers.

        **********************************************************************/

        protected struct Sprintf
        {
                Style           style;
                char[80]        local;
                char[]          buffer;
                char*           p, limit;

                void opCall (char[] dst, char[] workspace = null, DblFormat df = null)
                {
                        style (&utf8, workspace.length ? workspace : local, df);
                        p = buffer = dst;
                        limit = p + buffer.length;
                }

                int utf8 (char[] s)
                {
                        int len = s.length;
                        if (p+len < limit)
                           {
                           p[0..len] = s[0..len];
                           p += len;       
                           return len;
                           }
                        error ("Sprintf overflow");
                }

                int format (char[] fmt, ...)
                {
                        style.setFormat (fmt);
                        return print (style, _arguments, _argptr);
                }

                int format (char[] fmt, TypeInfo[] ti, va_list args)
                {
                        style.setFormat (fmt);
                        return print (style, ti, args);
                }
        }

        /***********************************************************************
        
                General purpose va_arg format routine, used by a number of
                classes and structs to emit printf() styled text. 

                This implementation simply converts its arguments into the
                appropriate internal type, and invokes the standard Mango
                type-formatter for each.

                Note that this can handle type-arrays in addition to the
                usual char[] ~ e.g. one can print an array of integers just
                as easily as a single integer.

                See Styled.parse() for the list of format specifiers.

        ***********************************************************************/

        static int print (inout Style style, TypeInfo[] arguments, va_list argptr)
        {      
                // internal method to map data types 
                uint getType (uint type)
                {
                        static  ubyte xlate[] = 
                                [
                                Type.Utf8, Type.Bool, 0, Type.Double, Type.Real, 
                                Type.Float, Type.Byte, Type.UByte, Type.Int, 0, 
                                Type.UInt, Type.Long, Type.ULong, 0, 0, 0, 0, 0, 
                                Type.Short, Type.UShort, Type.Utf16, 0, Type.Utf32, 
                                ];

                        if (type >= 'a' && type <= 'w')
                            type = xlate[type - 'a'];
                        else
                           if (type == 'P')
                               type = Type.Pointer;
                           else
                              error ("unknown type handed to Format.print()");
                        return type;
                }


                int length;

                // loop over each argument ...
                foreach (TypeInfo ti; arguments)
                        {
                        int count = 1;
                        void *p = argptr;

                        uint type = ti.classinfo.name[9];

                        // is this an array type?
                        if (type == 'A')
                           {
                           type = ti.classinfo.name[10];
                           p = (*(cast(void[]*) argptr)).ptr;
                           count = (*(cast(void[]*) argptr)).length;
                           }                

                        type = getType (type);
                        int width = Type.widths[type];
                        length += print (style, p, width * count, type);

                        argptr += (count > 1) ? (void[]).sizeof : ((width + int.sizeof - 1) & ~(int.sizeof - 1));
                        }

                // flush any remaining format text
                if (style.meta.length)
                   {
                   length += style.utf8 (style.meta);
                   style.meta = null;
                   }

                return length;
        }


        /***********************************************************************
        
                Convert various types into textual representation. This 
                is used by DisplayWriter/PrintWriter, and by the various
                users of the variadic-argument functions.

                See print() above, along with the DisplayWriter class.

                Note that print() can handle type-arrays in addition to the
                usual char[]; e.g. one can print an array of integers just
                as easily as a single integer. To support this, a new style
                flag has been introduced to terminate the format string at
                the point of use. That is, the rest of the format string is
                considered to be part of the current format specifier, and
                will be repeated for each array element. For example, the
                format string " %@x," has a preceeding space, a trailing
                comma, and the array flag '@'. This will output an array of 
                numeric values (char, byte, short, int, long, float, double,
                real, pointer) as a set of formatted hexadecimal strings.

                See Styled.parse() for the list of format specifiers.

        ***********************************************************************/

        static int print (inout Style style, void* src, uint bytes, uint type)
        {
                uint    iValue;
                ulong   lValue;
                double  fValue;
                uint    length;

                // convert format segment to Style
                if (style.parse (style.meta) == 0)
                    style.type = DefaultStyle[type];

                if (style.flags & Flags.Array)
                    style.meta = null;
                else
                   {
                   // flip remaining format, if array not specified
                   style.meta = style.tail;
                   style.tail = null;
                   }

                // get width of elements (note: does not work for bit[])
                int width = Type.widths[type];

                // for all bytes in source ...
                while (bytes)
                      {
                      switch (type)
                             {
                             case Type.Bool:
                                  iValue = *cast(bool*) src;
                                  if (style.type != 's')
                                      goto int32Format;
                                  length += style.utf8 (iValue ? "true" : "false");
                                  break;

                             case Type.Byte:
                                  iValue = *cast(byte*) src;
                                  goto int32Format;

                             case Type.UByte:
                                  iValue = *cast(ubyte*) src;
                                  goto int32Format;

                             case Type.Short:
                                  iValue = *cast(short*) src;
                                  goto int32Format;

                             case Type.UShort:
                                  iValue = *cast(ushort*) src;
                                  goto int32Format;

                             case Type.Int:
                             case Type.UInt:
                             case Type.Pointer:
int32:
                                  iValue = *cast(uint*) src;
int32Format:
                                  length += Int.format (iValue, style);
                                  break;

                             case Type.Long:
                             case Type.ULong:
int64:
                                  lValue = *cast(ulong*) src;
int64Format:
                                  length += Long.format (lValue, style);
                                  break;

                             case Type.Float:
                                  if (style.type == 'x' || 
                                      style.type == 'X')
                                      goto int32;
                                  fValue = *cast(float*) src;
                                  goto floating;

                             case Type.Double:
                                  if (style.type == 'x' || 
                                      style.type == 'X')
                                      goto int64;
                                  fValue = *cast(double*) src;
floating:
                                  if (style.dFormat == null)
                                      error ("floating point formatting not configured");
                                  length += style.dFormat (fValue, style);
                                  break;

                             case Type.Real:
                                  fValue = *cast(real*) src;
                                  goto floating;

                             case Type.Utf8:
                                  char[] s;
                                  if (style.type == 's')
                                     {
                                     int len = bytes;
                                     if (style.flags & Flags.Prec)
                                         if (style.precision < len)
                                             len = style.precision;
                                     s = cast(char[]) src[0..len];
                                     bytes = 1;
                                     }
                                  else
                                     if (style.type == 'c')
                                         s = cast(char[]) src[0..1];
                                     else
                                        {
                                        type = Type.UByte;
                                        continue;
                                        }
                                  length += style.emit (s);
                                  break;

                             default:
                                  error ("unknown type handed to Format.print()");
                             }

                      // bump counters and loop around for next instance
                      bytes -= width;
                      src += width;
                      }
                return length;
        }
}
