/*******************************************************************************

        @file Int.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        
        @version        Initial version, Feb 2005
        @author         Kris


*******************************************************************************/

module mango.format.Int;

private import mango.format.Number;

/******************************************************************************

        A set of functions for converting between string and integer 
        values. Additional methods are inherited via the super-class
        Number.

        While these functions are all static, they are encapsulated within 
        a class inheritance to preserve some namespace cohesion. One might 
        use structs for encapsualtion instead, but then inheritance would 
        be lost. Note that the root class, Styled, is abstract to prevent 
        accidental instantiation of these classes.

******************************************************************************/

class Int : Number
{       
        alias Number.format format;

        /**********************************************************************

                Format an integer based upon the Style provided. This
                is used by the Format class to convert integer values.
                See classes Styled & Format for more information.

        **********************************************************************/

        final static int format (int i, inout Style style)
        {
                return style.emit (format (style.workspace[0..style.width], i, cast(NFormat) style.type, style.flags));
        }


        /**********************************************************************

                Convert an integer using the provided NFormat and flags.

                The traditional printf() conversion specifiers are adhered 
                to, and the following types are supported:

                u - unsigned decimal
                d - signed decimal
                o - octal
                x - lowercase hexadecimal
                X - uppercase hexadecimal
                b - binary

                Modifiers supported include:

                #      : prefix the conversion with a type identifier
                +      : prefix positive decimals with a '+'
                space  : prefix positive decimals with one space
                0      : left-pad the number with zeros

                These modifiers are specifed via the 'flags' provided, 
                and are represented via these identifiers:

                #     : Flags.Hash
                +     : Flags.Plus
                space : Flags.Space
                0     : Flags.Zero

        **********************************************************************/

        final static char[] format (int i, NFormat fmt = NFormat.Decimal, Flags flags = 0)
        {
                return format (new char[32], i, fmt, flags);
        }


        /**********************************************************************

                Parse an integer value from the provided 'src' string. 
                The 'radix' used defaults to 10, but can specified as
                desired. In particular, one should specify a radix of
                zero to parse that also (via a type specifier).

        **********************************************************************/

        final static int parse (char[] src, uint radix=10, uint* ate=null)
        {
                bool    sign;
                uint    value;
                uint    count = trim (src, sign, &radix);

                foreach (char c; src [count..length])
                        {
                        if (c >= '0' && c <= '9')
                           {}
                        else
                           if (c >= 'a' && c <= 'f')
                               c -= 39;
                           else
                              if (c >= 'A' && c <= 'F')
                                  c -= 7;
                              else
                                 break;

                        value = value * radix + (c - '0');
                        ++count;
                        }

                if (ate)
                    *ate = count;
                return sign ? -value : value;
        }
}

