/*******************************************************************************

        @file Number.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        
        @version        Initial version; Feb 2005

        @author         Kris


*******************************************************************************/

module mango.format.Number;

private import  mango.format.Styled;

/******************************************************************************

        Utility functions for integer conversion.

        While these functions are all static, they are encapsulated within 
        a class inheritance to preserve some namespace cohesion. One might 
        use structs for encapsualtion instead, but then inheritance would 
        be lost. Note that the root class, Styled, is abstract to prevent 
        accidental instantiation of these classes.

******************************************************************************/

class Number : Styled
{
        enum NFormat:ubyte {Binary='b', Octal='o', Hex='x', HexUpper='X', Decimal='d', Unsigned='u'};

        /***********************************************************************
        
                Format numeric values into the provided output buffer. The
                traditional printf() conversion specifiers are adhered to,
                and the following types are supported:

                u - unsigned decimal
                d - signed decimal
                o - octal
                x - lowercase hexadecimal
                X - uppercase hexadecimal
                b - binary

                Modifiers supported include:

                #      : prefix the conversion with a type identifier
                +      : prefix positive decimals with a '+'
                space  : prefix positive decimals with one space
                0      : left-pad the number with zeros

                These modifiers are specifed via the 'flags' provided, 
                and are represented via these identifiers:

                #     : Flags.Hash
                +     : Flags.Plus
                space : Flags.Space
                0     : Flags.Zero

        ***********************************************************************/

        final static char[] format (char[] dst, long i, NFormat fmt = NFormat.Decimal, Flags flags = 0)
        {
                char[]  prefix;
                int     len = dst.length;
                   
                // must have some buffer space to operate within! 
                if (len)
                   {
                   uint    radix;
                   char[]  numbers = "0123456789abcdef";

                   // pre-conversion setup
                   switch (fmt)
                          {
                          default:
                          case NFormat.Decimal:
                               if (i < 0)
                                  {
                                  prefix = "-";
                                  i = -i;
                                  }
                               else
                                  if (flags & Flags.Space)
                                      prefix = " ";
                                  else
                                     if (flags & Flags.Plus)
                                         prefix = "+";
                               // fall through!
                          case NFormat.Unsigned:
                               radix = 10;
                               break;

                          case NFormat.Binary:
                               radix = 2;
                               if (flags & Flags.Hash)
                                   prefix = "0b";
                               break;

                          case NFormat.Octal:
                               radix = 8;
                               if (flags & Flags.Hash)
                                   prefix = "0o";
                               break;

                          case NFormat.Hex:
                               radix = 16;
                               if (flags & Flags.Hash)
                                   prefix = "0x";
                               break;

                          case NFormat.HexUpper:
                               radix = 16;
                               numbers = "0123456789ABCDEF";
                               if (flags & Flags.Hash)
                                   prefix = "0X";
                               break;
                          }
        
                   // convert number to text
                   char* p = dst.ptr + len;
                   if (i <= uint.max)
                      {
                      uint v = i;
                      do {
                         *--p = numbers[v % radix];
                         } while ((v /= radix) && --len);
                      }
                   else
                      {
                      ulong v = i;
                      do {
                         *--p = numbers[v % radix];
                         } while ((v /= radix) && --len);
                      }
                   }

                // are we about to overflow?
                if (--len < 0 || 0 > (len -= prefix.length))
                    error ("output buffer too small in Number.format()");

                // prefix number with zeros? 
                if (flags & Flags.Zero)
                   {
                   dst [prefix.length .. len + prefix.length] = '0';
                   len = 0;
                   }
                
                // write optional prefix string ...
                dst [len .. len + prefix.length] = prefix[];

                // return slice of provided output buffer
                return dst [len .. dst.length];                               
        } 


        /**********************************************************************
        
                Strip leading whitespace, extract an optional numeric sign,
                and optionally strip and return a possible radix prefix.

                This can be used as a precursor to the conversion of digits
                into a number.

                Returns the number of matching characters.

        **********************************************************************/

        package static uint trim (char[] digits, inout bool sign, uint* radix=null)
        {
                char   c;
                char*  p = digits;
                int    len = digits.length;

                // strip off whitespace and sign characters
                for (c = *p; len; c = *++p, --len)
                     if (c == ' ' || c == '\t')
                        {}
                     else
                        if (c == '-')
                            sign = true;
                        else
                           if (c == '+')
                               sign = false;
                           else
                              break;

                // strip off a radix specifier also?
                if (radix && *radix == 0)
                    if (c == '0' && len)
                        switch (*++p)
                               {
                               case 'x':
                               case 'X':
                                    ++p;
                                    *radix = 16;
                                    break;

                               case 'b':
                               case 'B':
                                    ++p;
                                    *radix = 2;
                                    break;

                               case 'o':
                               case 'O':
                                    ++p;
                               default:
                                    *radix = 8;
                                    break;
                               } 
                    else
                       *radix = 10;

                // return number of characters eaten
                return p - digits.ptr;
        }
}


