/*******************************************************************************

        @file Styled.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        
        @version        Initial version, Feb 2005
        @author         Kris


*******************************************************************************/

module mango.format.Styled;

version (Ares)
         private import  std.c.ctype;
      else
         private import  std.ctype;

private import  mango.io.Exception;

/******************************************************************************

        Root class of the formatting hierarchy.

        While these functions are all static, they are encapsulated within 
        a class inheritance to preserve some namespace cohesion. One might 
        use structs for encapsualtion instead, but then inheritance would 
        be lost. Note that this root class is abstract to prevent accidental 
        instantiation of subclasses.

        See Styled.parse() for format specifiers.

******************************************************************************/

abstract class Styled
{
        private typedef int delegate (char[]) Utf8Sink;
        private typedef int function (double, inout Style) DblFormat;
        
        // array of spaces to pad styled-output
        private static char[256] Spaces = ' ';

        /**********************************************************************

                Declare stylistic flags 

        **********************************************************************/

        enum Flags 
        {
                Fill   = 1,                     // do some kind of padding
                Left    = Fill << 1,            // left justify
                Prec	= Left << 1,            // precision was provided
                Hash	= Prec << 1,            // prefix integer with type
                Space	= Hash << 1,            // prefix with space
                Zero	= Space << 1,           // prefix integer with zero
                Sign	= Zero << 1,            // unused
                Comma	= Sign << 1,            // unused
                Plus	= Comma << 1,           // prefix decimal with '+'
                Array	= Plus << 1,            // array flag
        };

        /**********************************************************************

                The state of a stylistic conversion. This can be embedded
                within a class, or instantiated on the stack. Be sure to
                initialize the Style via methods opCall() & setFormat()

        **********************************************************************/

        struct Style
        {
                Utf8Sink        utf8;
                int             type,
                                width,
                                precision;
                Flags           flags;
                char[]          head,
                                tail,
                                meta;
                DblFormat       dFormat;
                char[]          workspace;  

                /**************************************************************

                        Configure this Style with an output handler, a
                        workspace area, and a floating point handler.
                        The latter is optional.

                **************************************************************/

                void opCall (Utf8Sink utf8, char[] workspace, DblFormat dFormat = null)
                {
                        this.utf8 = utf8;
                        this.dFormat = dFormat;
                        this.workspace = workspace;
                }

                /**************************************************************

                        Set the format string for this Style

                **************************************************************/

                void setFormat (char[] format)
                {
                        meta = format;
                }

                /**************************************************************

                        Clear the current state. This is typically used
                        internally only.

                **************************************************************/

                void clear ()
                {
                        flags = 0;
                        head = tail = null;   
                        width = workspace.length;
                }
        
                /**************************************************************

                        Emit a field, surrounded by optional prefix and
                        postfix strings, and optionally padded with spaces.

                **************************************************************/

                int emit (char[] field)
                {
                        int i = 0;
                        int pad = 0;

                        // emit prefix?
                        if (head.length)
                            i += utf8 (head);                   

                        // should we pad output?
                        if (flags & Flags.Fill && flags & Flags.Space)
                           {
                           pad = width - field.length;
                           if (pad < 0)
                               pad = 0;
                           else
                              if (pad > Spaces.length)
                                  error ("space padding greater than 256");

                           // right-aligned?
                           if (flags & Flags.Left == 0)
                              {
                              i += utf8 (Spaces[0..pad]);
                              pad = 0;
                              }                        
                           }
                           
                        // emit field itself
                        i += utf8 (field);

                        // any trailing padding?
                        if (pad)
                            i += utf8 (Spaces[0..pad]);

                        // emit postfix
                        if (tail.length)
                            i += utf8 (tail);   
                                            
                        return i;
                }

                /**************************************************************

                        Parse a format specifier into its constituent
                        flags and values. Syntax follows the traditional
                        printf() approach, as follows:

                        %[flags][width][.precision]type

                        Where 'type' is one of:

                        s : string format
                        c : character format
                        d : signed format
                        u : unsigned format
                        x : hexadecimal format
                        X : uppercase hexadecimal format
                        e : scientific notation 
                        f : floating point format
                        g : 'e' or 'f', based upon width

                        Note that there are no variants on the format
                        types ~ long, int, short, and byte differences
                        are all handled internally.
                        
                        The 'flags' supported:

                        space : prefix negative integer with one space;
                                pad any type when combined with a width
                                specifier
                        -     : left-align fields padded with spaces
                        +     : prefix positive integer with one '+'
                        0     : prefix integers with zeroes; requires a
                                width specification
                        #     : prefix integers with a type specifier
                        @     : Array specifier

                        The 'width' should be specified for either zero or
                        space padding, and may be used with all formatting
                        types.

                        A 'precision' can be used to stipulate the number
                        of decimal-places, or a slice of a text string.

                        Note that the Format package supports array-output
                        in addition to the usual printf() output.

                **************************************************************/

                char parse (char[] format)
                {
                        clear ();
                        head = format;
                        char* p = format.ptr;

                        for (int i = format.length; --i > 0; ++p)
                             if (*p == '%')
                                 if (p[1] == '%')
                                     ++p;
                                 else
                                    {
                                    int len = p - format.ptr;
                                    head = format [0..len];   
                                    while (1)
                                          {
                                          switch (--i, *++p)
                                                 {
                                                 case '-': 
                                                      flags |= Flags.Left;
                                                      continue;

                                                 case '+': 
                                                      flags |= Flags.Plus;
                                                      continue;

                                                 case '#': 
                                                      flags |= Flags.Hash;
                                                      continue;

                                                 case ' ': 
                                                      flags |= Flags.Space;
                                                      continue;

                                                 case '0': 
                                                      flags |= Flags.Zero;
                                                      continue;

                                                 case '@': 
                                                      flags |= Flags.Array;
                                                      continue;

                                                 default: 
                                                      ++i;
                                                      break;
                                                 }
                                          break;
                                          }

                                    if (isdigit(*p))
                                       {
                                       int tmp;
                                       do {
                                          tmp = tmp * 10 + (*p - '0');
                                          } while (--i && isdigit(*++p));

                                       flags |= Flags.Fill;
                                       width = tmp;
                                       }
                                    else
                                       flags &= ~Flags.Zero;


                                    if (*p == '.')
                                       {
                                       int tmp;
                                       while (i-- && isdigit(*++p))
                                              tmp = tmp * 10 + (*p - '0');

                                       flags |= Flags.Prec;
                                       precision = tmp;
                                       }

                                    if (--i < 0)
                                        error ("missing format specifier");

                                    tail = format [format.length-i..format.length];
                                    return type = *p;
                                    }

                        return type = 0;
                }
        }


        /**********************************************************************

                Throw an exception with the provided messsage

        **********************************************************************/

        final static void error (char[] msg)
        {
                throw new FormatException (msg);
        }
}


/*******************************************************************************

        This exception is thrown by the Format package if it finds issue
        with the format specifiers and so on.

*******************************************************************************/

class FormatException : Exception
{
        /***********************************************************************
        
                Construct exception with the provided text string

        ***********************************************************************/

        this (char[] msg)
        {
                super (msg);
        }
}

