/*******************************************************************************

        @file HttpClient.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, April 2004      
        @author         Kris


*******************************************************************************/

module mango.http.client.HttpClient;

private import  mango.sys.System;

private import  mango.utils.Text;

private import  mango.format.Int;

private import  mango.io.Uri,
                mango.io.Token,
                mango.io.Buffer,
                mango.io.SocketConduit,
                mango.io.Exception,
                mango.io.Tokenizer,
                mango.io.DisplayWriter;

private import  mango.http.server.HttpParams,  
                mango.http.server.HttpHeaders,
                mango.http.server.HttpCookies,
                mango.http.server.HttpResponse;

/*******************************************************************************

        Supports the basic needs of a client making requests of an HTTP
        server. The following is an example of how this might be used:

        @code
        // callback for client reader
        void sink (char[] content)
        {
                Stdout.put (content);
        }

        // create client for a GET request
        auto HttpClient client = new HttpClient (HttpClient.Get, "http://www.digitalmars.com/d/intro.html");

        // setup a Host header
        client.getRequestHeaders.add (HttpHeader.Host, client.getUri.getHost());

        // make request
        client.open ();

        // check return status for validity
        if (client.isResponseOK)
           {
           // extract content length (be aware of -1 return, for no header)
           int length = client.getResponseHeaders.getInt (HttpHeader.ContentLength);
           if (length < 0)
               length = int.max;
        
           // display all returned headers
           Stdout.put (client.getResponseHeaders);
        
           // display remaining content
           client.read (&sink, length);
           }
        else
           Stderr.put (client.getResponse);
        @endcode

*******************************************************************************/

class HttpClient
{       
        // this is struct rather than typedef to avoid compiler bugs
        private struct RequestMethod
        {
                final char[]            name;
        }    
                        
        // class members; there's a surprising amount of stuff here!
        private MutableUri              uri;
        private IBuffer                 input,
                                        output;
        private SocketConduit           socket;
        private RequestMethod           method;
        private InternetAddress         address;
        private HttpMutableParams       paramsOut;
        private HttpHeaders             headersIn;
        private HttpMutableHeaders      headersOut;
        private HttpMutableCookies      cookiesOut;
        private ResponseLine            responseLine;

        // default to three second timeout on read operations ...
        private static final uint DefaultReadTimeout = System.Interval.Second * 3;

        // use HTTP v1.0 ?
        private static const char[] DefaultHttpVersion = "HTTP/1.1";

        // standard set of request methods ...
        static const RequestMethod      Get = {"GET"},
                                        Put = {"PUT"},
                                        Head = {"HEAD"},
                                        Post = {"POST"},
                                        Trace = {"TRACE"},
                                        Delete = {"DELETE"},
                                        Options = {"OPTIONS"},
                                        Connect = {"CONNECT"};

        /***********************************************************************
        
                Create a client for the given URL. The argument should be
                fully qualified with an "http:" or "https:" scheme, or an
                explicit port should be provided.

        ***********************************************************************/

        this (RequestMethod method, char[] url)
        {
                this (method, new MutableUri(url));
        }

        /***********************************************************************
        
                Create a client with the provided Uri instance. The Uri should 
                be fully qualified with an "http:" or "https:" scheme, or an
                explicit port should be provided. 

        ***********************************************************************/

        this (RequestMethod method, MutableUri uri)
        {
                this.uri = uri;
                this.method = method;

                responseLine = new ResponseLine ();
                headersIn    = new HttpHeaders  ();

                paramsOut    = new HttpMutableParams  (new Buffer (1024 * 1));
                headersOut   = new HttpMutableHeaders (new Buffer (1024 * 4));
                cookiesOut   = new HttpMutableCookies (headersOut);

                // decode the host name (may take a second or two)
                address = new InternetAddress (uri.getHost(), uri.getValidPort());
        }

        /***********************************************************************
        
                Attempt to clean up when garbage collected

        ***********************************************************************/

        ~this ()
        {
                close ();
        }

        /***********************************************************************
        
                Get the current input headers, as returned by the host request.

        ***********************************************************************/

        HttpHeaders getResponseHeaders()
        {
                return headersIn;
        }

        /***********************************************************************
        
                Gain access to the request headers. Use this to add whatever
                headers are required for a request. 

        ***********************************************************************/

        HttpMutableHeaders getRequestHeaders()
        {
                return headersOut;
        }

        /***********************************************************************
        
                Gain access to the request parameters. Use this to add x=y
                style parameters to the request. These will be appended to
                the request assuming the original Uri does not contain any
                of its own.

        ***********************************************************************/

        HttpMutableParams getRequestParams()
        {
                return paramsOut;
        }

        /***********************************************************************
        
                Return the Uri associated with this client

        ***********************************************************************/

        Uri getUri()
        {
                return uri;
        }

        /***********************************************************************
        
                Return the response-line for the latest request. This takes 
                the form of "version status reason" as defined in the HTTP
                RFC.

        ***********************************************************************/

        ResponseLine getResponse()
        {
                return responseLine;
        }

        /***********************************************************************
        
                Return the HTTP status code set by the remote server

        ***********************************************************************/

        int getStatus()
        {
                return responseLine.getStatus();
        }

        /***********************************************************************
        
                Return whether the response was OK or not

        ***********************************************************************/

        bool isResponseOK()
        {
                return getStatus() == HttpResponseCode.OK;
        }

        /***********************************************************************
        
                Add a cookie to the outgoing headers

        ***********************************************************************/

        void addCookie (Cookie cookie)
        {
                cookiesOut.add (cookie);
        }

        /***********************************************************************
        
                Close all resources used by a request. You must invoke this 
                between successive open() calls.

        ***********************************************************************/

        void close ()
        {
                if (socket)
                   {
                   socket.shutdown ();
                   socket.close ();
                   socket = null;
                   }
        }

        /***********************************************************************

                Reset the client such that it is ready for a new request.
        
        ***********************************************************************/

        void reset ()
        {
                headersIn.reset();
                headersOut.reset();
                paramsOut.reset();
        }

        /***********************************************************************

                Overridable method to create a Socket. You may find a need 
                to override this for some purpose; perhaps to add input or 
                output filters.
                 
        ***********************************************************************/

        protected SocketConduit createSocket ()
        {
                return new SocketConduit;
        }

        /***********************************************************************
        
                Make a request for the resource specified via the constructor,
                using a callback for pumping additional data to the host. This 
                defaults to a three-second timeout period. The return value 
                represents the input buffer, from which all returned headers 
                and content may be accessed.
                
        ***********************************************************************/

        IBuffer open (IWritable pump)
        {
                return open (DefaultReadTimeout, pump);
        }

        /***********************************************************************
        
                Make a request for the resource specified via the constructor,
                using the specified timeout period (in milli-seconds).The 
                return value represents the input buffer, from which all
                returned headers and content may be accessed.
                
        ***********************************************************************/

        IBuffer open (uint timeout = DefaultReadTimeout)
        {
                return open (timeout, null);
        }

        /***********************************************************************
        
                Make a request for the resource specified via the constructor
                using the specified timeout period (in micro-seconds), and a
                user-defined callback for pumping additional data to the host.
                The callback would be used when uploading data during a 'put'
                operation (or equivalent). The return value represents the 
                input buffer, from which all returned headers and content may 
                be accessed.
                
        ***********************************************************************/

        IBuffer open (uint timeout, IWritable pump)
        {
                IWriter emit;
                
                // create socket, and connect it 
                socket = createSocket;
                socket.setTimeout (timeout);
                socket.connect (address);
                
                // create buffers for input and output
                input = socket.createBuffer;
                output = socket.createBuffer;
                  
                // bind a writer to the socket
                emit = new DisplayWriter (output);

                // attach/extend query parameters if user has added some
                char[] query = uri.extendQuery (paramsOut.toOutputString);

                // format request 
                emit (method.name)
                     (' ')
                     (uri.getPath);

                if (query.length)
                    emit ('?') (query);

                emit (' ')
                     (DefaultHttpVersion)
                     (CR)
                     (headersOut)
                     (CR);

                // user has more data to send?
                if (pump)
                    pump.write (emit);
                
                // send entire request
                emit.flush ();

                // Token for initial parsing of input header lines
                CompositeToken line = new CompositeToken (Tokenizers.line, input);

                // skip any blank lines
                while (line.next() && line.getLength() == 0) 
                      {}

                // is this a bogus request?
                if (input.readable == 0)
                    throw new IOException ("truncated response");
                    
                // read response line
                responseLine.parse (line.toString);

                // parse headers and go home
                headersIn.parse (input);
                return input;
        }

        /***********************************************************************
        
        ***********************************************************************/

        void read (void delegate (char[]) sink, long length = long.max)
        {
                // display remaining content
                do {
                   length -= input.readable;
                   sink (input.toString);
                   input.clear ();
                   } while (length > 0 && socket.read(input) != socket.Eof);
        }
}


/******************************************************************************

        Class to represent an HTTP response-line

******************************************************************************/

private class ResponseLine : IWritable
{
        private Buffer          buf;
        private BoundToken      token;
        private char[]          reason,
                                status,
                                vershion;

        /**********************************************************************

                Construct response-line with pre-allocated buffer & token

        **********************************************************************/

        this ()
        {
                buf = new Buffer;

                // bind tokens to a space-tokenizer
                token = new BoundToken (new SpaceTokenizer());
        }

        /**********************************************************************

                Parse the the given response-line into its constituent 
                components.

        **********************************************************************/

        void parse (char[] line)
        {
                // setup our buffer with the request
                buf.setValidContent (line);

                if (token.next (buf))
                   {
                   vershion = token.toString();
                   if (token.next (buf))
                      {
                      status = token.toString();
                      reason = line[buf.getPosition..line.length];
                      return;
                      }
                   }
               throw new IOException ("Invalid HTTP response-line: '"~line~"'");
        }

        /**********************************************************************

                Return HTTP version

        **********************************************************************/

        char[] getVersion ()
        {
                return vershion;
        }

        /**********************************************************************

                Return reason text

        **********************************************************************/

        char[] getReason ()
        {
                return reason;
        }

        /**********************************************************************

                Return status integer

        **********************************************************************/

        int getStatus ()
        {
                return Int.parse (status);
        }

        /**********************************************************************

                convert back to original string

        **********************************************************************/

        override char[] toString ()
        {
                return vershion~" "~status~" "~reason;
        }

        /**********************************************************************

                Output the string via the given writer

        **********************************************************************/

        void write (IWriter writer)
        {
               writer.put(toString).cr();
        }
}


