/*******************************************************************************

        @file Buffer.d

        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      
        @version        Initial version; March 2004

        @author         Kris


*******************************************************************************/

module mango.io.Buffer;

private import  mango.io.Exception;

private import  mango.io.model.IConduit;

public  import  mango.io.model.IBuffer;

/******************************************************************************

******************************************************************************/

extern (C)
{
        void * memcpy (void *dst, void *src, uint);
}       

/*******************************************************************************

        The basic premise behind this IO package is as follows:

        @li the central concept is that of a buffer. The buffer acts
           as a queue (line) where items are removed from the front
           and new items are added to the back. Buffers are modeled 
           by mango.io.model.IBuffer, and a concrete implementation is 
           provided this class.

        @li buffers can be written to directly, but a Reader and/or
           Writer are typically used to read & write formatted data.
           These readers & writers are bound to a specific buffer;
           often the same buffer. It's also perfectly legitimate to 
           bind multiple writers to the same buffer; they will all
           behave serially as one would expect. The same applies to
           multiple readers on the same buffer. Readers and writers
           support three styles of IO: put/get, the C++ style << &
           >> operators, and the () whisper style. All operations 
           can be chained.

        @li Any class can be made compatable with the reader/writer
           framework by implementing the IReadable and/or IWritable 
           interfaces. Each of these specify just a single method.

        @li Buffers may also be tokenized. This is handy when one is
           dealing with text input, and/or the content suits a more
           fluid format than most typical readers & writers support.
           Tokens are mapped directly onto buffer content, so there
           is only minor overhead in using them. Tokens can be read
           and written by reader/writers also, using a more relaxed
           set of rules than those applied to integral IO.

        @li buffers are sometimes memory-only, in which case there
           is nothing left to do when a reader (or tokenizer) hits
           end of buffer conditions. Other buffers are themselves 
           bound to a Conduit. When this is the case, a reader will 
           eventually cause the buffer to reload via its associated 
           conduit. Previous buffer content will thus be lost. The
           same concept is applied to writers, whereby they flush 
           the content of a full buffer to a bound conduit before 
           continuing. 

        @li conduits provide virtualized access to external content,
           and represent things like files or Internet connections.
           They are just a different kind of stream. Conduits are
           modelled by mango.io.model.IConduit, and implemented via
           classes FileConduit and SocketConduit. Additional kinds
           of conduit are easy to construct: one either subclasses
           mango.io.Conduit, or implements mango.io.model.IConduit. 
           A conduit reads and writes from/to a buffer in big chunks
           (typically the entire buffer).


        An example of how to append a buffer follows:

        @code
        char[] foo = "to write some D";

        // create a small buffer
        Buffer buf = new Buffer (256);

        // append some text directly to it
        buf.append("now is the time for all good men ").append(foo);

        // output the combined string
        Stdout (buf.toString) (CR);
        @endcode

        You might use a GrowableBuffer instead where you wish to 
        append beyond a preset limit. One common usage of buffers 
        is in conjunction with a conduit, such as FileConduit. Each 
        Conduit exposes a factory for creating a buffer of the most
        appropriate size or flavour:

        @code
        FileConduit fc = new FileConduit ("file.name");
        Buffer buf = fc.createBuffer ();
        @endcode

        However, this is typically hidden by higher level constructors 
        such as those of Reader and Writer derivitives. For example:

        @code
        FileConduit fc = new FileConduit ("file.name");
        Reader reader = new Reader (fc);
        @endcode

        There is indeed a buffer between the Reader and Conduit, but 
        explicit construction is unecessary in many cases.  See both 
        Reader and Writer for examples of formatted IO.

*******************************************************************************/

class Buffer : IBuffer
{
        private void[]          data;
        private uint            limit;
        private uint            capacity;
        private uint            position;
        private IConduit        conduit;

        // error messages
        protected static char[] overflow  = "output buffer overflow";
        protected static char[] underflow = "input buffer underflow";
        protected static char[] eofRead   = "end-of-file whilst reading";
        protected static char[] eofWrite  = "end-of-file whilst writing";


        /***********************************************************************
        
                Ensure the buffer remains valid between method calls
                 
        ***********************************************************************/

        invariant 
        {
               assert (position <= limit);
               assert (limit <= capacity);
        }

        /***********************************************************************
        
                Construct a Buffer with the specified number of bytes.

        ***********************************************************************/

        this (uint capacity = 0)
        {
                this (new ubyte[capacity]);
        }

        /***********************************************************************
        
                Prime buffer with an application-supplied array. There is 
                no readable data present, and writing begins at position 0.

        ***********************************************************************/

        this (void[] data)
        {
                this (data, 0);
        }

        /***********************************************************************
        
                Prime buffer with an application-supplied array, and 
                indicate how much readable data is already there. A
                write operation will begin writing immediately after
                the existing content.

        ***********************************************************************/

        this (void[] data, uint readable)
        {
                setContent (data, readable);
        }

        /***********************************************************************
        
                Throw an exception with the provided message

        ***********************************************************************/

        protected void error (char[] msg)
        {
                throw new IOException (msg);
        }

        /***********************************************************************
                
                Create an instance of an IBuffer. Use this when you
                don't know anything about the concrete implementation,
                and have only the IBuffer interface available

                Returns a Buffer with no content.

        ***********************************************************************/

        IBuffer create ()
        {
                return new Buffer;
        }

        /***********************************************************************
                
                Return the backing array

        ***********************************************************************/

        void[] getContent ()
        {
                return data;
        }

        /***********************************************************************
        
                Set the backing array with all content readable. Writing
                to this will either flush it to an associated conduit, or
                raise an Eof condition. Use IBuffer.clear() to reset the
                content (make it all writable).

        ***********************************************************************/

        IBuffer setValidContent (void[] data)
        {
                return setContent (data, data.length);
        }

        /***********************************************************************
        
                Set the backing array with some content readable. Writing
                to this will either flush it to an associated conduit, or
                raise an Eof condition. Use IBuffer.clear() to reset the
                content (make it all writable).

        ***********************************************************************/

        IBuffer setContent (void[] data, uint readable)
        {
                this.data = data;
                this.limit = readable;
                this.capacity = data.length;   

                // reset to start of input
                this.position = 0;

                return this;            
        }

        /***********************************************************************

                Overridable method to grow the buffer size when it becomes 
                full. Default is to not grow at all.

        ***********************************************************************/

        protected bool grow (uint size)
        {
                return false;
        }

        /***********************************************************************
        
                Bulk copy of data from 'src'. Limit is adjusted by 'size'
                bytes.

        ***********************************************************************/

        protected void copy (void *src, uint size)
        {
                data[limit..limit+size] = src[0..size];
                limit += size;
        }

        /***********************************************************************
        
                Read a chunk of data from the buffer, loading from the
                conduit as necessary. The specified number of bytes is
                loaded into the buffer, and marked as having been read 
                when the 'eat' parameter is set true. When 'eat' is set
                false, the read position is not adjusted.

                Returns the corresponding buffer slice when successful, 
                or null if there's not enough data available (Eof; Eob).

        ***********************************************************************/

        void[] get (uint size, bool eat = true)
        {   
                if (size > readable)
                   {
                   if (conduit is null)
                       error (underflow);

                   // make some space? This will try to leave as much content
                   // in the buffer as possible, such that entire records may
                   // be aliased directly from within.
                   if (size > writable)
                      {
                      if (size > capacity)
                          error (underflow);
                      compress ();
                      }

                   // populate tail of buffer with new content
                   do {
                      if (conduit.read (this) == IConduit.Eof)
                          error (eofRead);
                      } while (size > readable);
                   }

                uint i = position;
                if (eat)
                    position += size;
                return data [i .. i + size];               
        }

        /***********************************************************************
        
                Append an array of data to this buffer, and flush to the
                conduit as necessary. Returns a chaining reference if all 
                data was written; throws an IOException indicating eof or 
                eob if not.

                This is often used in lieu of a Writer.

        ***********************************************************************/

        IBuffer append (void[] src)        
        {               
                uint size = src.length;

                if (size > writable)
                    // can we write externally?
                    if (conduit)
                       {
                       flush ();

                       // check for pathological case
                       if (size > capacity)
                          {
                          // we could write directly to the conduit, but
                          // then we'd lose support for partial writes ...
                          void[] tmp = data;

                          setContent (src, size);
                          flush ();
                          setContent (tmp, 0);
                          return this;
                          }
                       }
                    else
                       // try to grow buffer ...
                       if (! grow (size))
                             error (overflow);

                copy (src, size);
                return this;
        }

        /***********************************************************************
        
                Return a char[] slice of the buffer up to the limit of
                valid content.

        ***********************************************************************/

        override char[] toString ()
        {       
                return cast(char[]) data[position..limit];
        }

        /***********************************************************************
        
                Skip ahead by the specified number of bytes, streaming from 
                the associated conduit as necessary.
        
                Can also reverse the read position by 'size' bytes. This may
                be used to support lookahead-type operations.

                Returns true if successful, false otherwise.

        ***********************************************************************/

        bool skip (int size)
        {
                if (size < 0)
                   {
                   size = -size;
                   if (position >= size)
                      {
                      position -= size;
                      return true;
                      }
                   return false;
                   }
                return get (size) !is null;
        }

        /***********************************************************************
        
                Return count of readable bytes remaining in buffer. This is 
                calculated simply as limit() - position()

        ***********************************************************************/

        uint readable ()
        {
                return limit - position;
        }               

        /***********************************************************************
        
                Return count of writable bytes available in buffer. This is 
                calculated simply as capacity() - limit()

        ***********************************************************************/

        uint writable ()
        {
                return capacity - limit;
        }               

        /***********************************************************************

                Exposes the raw data buffer at the current write position, 
                The delegate is provided with a void[] representing space
                available within the buffer at the current write position.

                The delegate should return the appropriate number of bytes 
                if it writes valid content, or IConduit.Eof on error.

                Returns whatever the delegate returns.

        ***********************************************************************/

        int write (int delegate (void[]) dg)
        {
                int count = dg (data [limit..capacity]);

                if (count != IConduit.Eof) 
                   {
                   limit += count;
                   assert (limit <= capacity);
                   }
                return count;
        }               

        /***********************************************************************

                Exposes the raw data buffer at the current read position. The
                delegate is provided with a void[] representing the available
                data, and should return zero to leave the current read position
                intact. 
                
                If the delegate consumes data, it should return the number of 
                bytes consumed; or IConduit.Eof to indicate an error.

                Returns whatever the delegate returns.

        ***********************************************************************/

        int read (int delegate (void[]) dg)
        {
                
                int count = dg (data [position..limit]);
                
                if (count != IConduit.Eof)
                   {
                   position += count;
                   assert (position <= limit);
                   }
                return count;
        }               

        /***********************************************************************

                If we have some data left after an export, move it to 
                front-of-buffer and set position to be just after the 
                remains. This is for supporting certain conduits which 
                choose to write just the initial portion of a request.
                
                Limit is set to the amount of data remaining. Position 
                is always reset to zero.

        ***********************************************************************/

        IBuffer compress ()
        {
                uint r = readable ();

                if (position > 0 && r > 0)
                    // content may overlap ...
                    memcpy (&data[0], &data[position], r);

                position = 0;
                limit = r;
                return this;
        }               

        /***********************************************************************
        
                flush the contents of this buffer to the related conduit.
                Throws an IOException on premature eof.

        ***********************************************************************/

        void flush ()
        {
                if (conduit)
                    while (readable)
                           if (conduit.write (this) == conduit.Eof)
                               error (eofWrite);
        }               

        /***********************************************************************
        
                Reset 'position' and 'limit' to zero

        ***********************************************************************/

        IBuffer clear ()
        {
                position = limit = 0;
                return this;
        }               

        /***********************************************************************
        
                Returns the limit of readable content within this buffer

        ***********************************************************************/

        uint getLimit ()
        {
                return limit;
        }

        /***********************************************************************
        
                Returns the total capacity of this buffer

        ***********************************************************************/

        uint getCapacity ()
        {
                return capacity;
        }

        /***********************************************************************
        
                Returns the current read-position within this buffer

        ***********************************************************************/

        uint getPosition ()
        {
                return position;
        }

        /***********************************************************************
        
                Returns the conduit associated with this buffer. Returns 
                null if the buffer is purely memory based; that is, it's
                not backed by some external medium.

                Buffers do not require an external conduit to operate, but 
                it can be convenient to associate one. For example, methods
                read and write use it to import/export content as necessary.

        ***********************************************************************/

        IConduit getConduit ()
        {
                return conduit;
        }

        /***********************************************************************
        
                Sets the external conduit associated with this buffer.

                Buffers do not require an external conduit to operate, but 
                it can be convenient to associate one. For example, methods
                read and write use it to import/export content as necessary.

        ***********************************************************************/

        void setConduit (IConduit conduit)
        {
                this.conduit = conduit;
        }
}


/*******************************************************************************

        Subclass to provide support for content growth. This is handy when
        you want to keep a buffer around as a scratchpad.

*******************************************************************************/

class GrowableBuffer : Buffer
{
        /***********************************************************************
        
                Create a GrowableBuffer with the specified initial size.

        ***********************************************************************/

        this (uint size)
        {
                super (size);
        }

        /***********************************************************************

                Overridable method to grow the buffer size when it becomes 
                full. Default is to not grow at all.
        
        ***********************************************************************/

        protected override bool grow (uint size)
        {
                if (data.length < uint.max / 2)
                   {
                   uint i = data.length * 2;
                   if (i < size)
                       i = size;
                   data.length = i;
                   capacity = i;
                   return true;
                   }
                return false;
        }
}



/*******************************************************************************

        Subclass to treat the buffer as a seekable entity, where all 
        capacity is available for reading and/or writing. To achieve 
        this we must effectively disable the 'limit' watermark, and 
        locate write operations around 'position' instead. 

*******************************************************************************/

class MappedBuffer : Buffer
{
        /***********************************************************************
                
                Construct an empty MappedBuffer 

        ***********************************************************************/

        this ()
        {
                super (0);
        }

        /***********************************************************************
                
                Close this mapped buffer

        ***********************************************************************/

        abstract void close ();

        /***********************************************************************
                
                Flush any dirty content out to the drive

        ***********************************************************************/

        abstract void flush ();

        /***********************************************************************
                
                Set the read/write position

        ***********************************************************************/

        void setPosition (uint position)
        {
                this.position = position;
        }

        /***********************************************************************
        
                Seek to the specified position within the buffer, and return
                the byte offset of the new location (relative to zero).

        ***********************************************************************/

        long seek (long offset, ISeekable.SeekAnchor anchor)
        {
                long pos = capacity;

                if (anchor == ISeekable.SeekAnchor.Begin)
                    pos = offset;
                else
                   if (anchor == ISeekable.SeekAnchor.End)
                       pos -= offset;
                   else
                      pos = position + offset;

                return position = pos;
        }

        /***********************************************************************
        
                Return count of writable bytes available in buffer. This is 
                calculated simply as capacity() - limit()

        ***********************************************************************/

        override uint writable ()
        {
                return capacity - position;
        }               

        /***********************************************************************
        
                Bulk copy of data from 'src'. Position is adjusted by 'size'
                bytes.

        ***********************************************************************/

        override protected void copy (void *src, uint size)
        {
                data[position..position+size] = src[0..size];
                position += size;
        }

        /***********************************************************************

                Exposes the raw data buffer at the current write position, 
                The delegate is provided with a void[] representing space
                available within the buffer at the current write position.

                The delegate should return the appropriate number of bytes 
                if it writes valid content, or IConduit.Eof on error.

                Returns whatever the delegate returns.

        ***********************************************************************/

        override int write (int delegate (void[]) dg)
        {
                int count = dg (data [position..capacity]);

                if (count != IConduit.Eof) 
                   {
                   position += count;
                   assert (position <= capacity);
                   }
                return count;
        }               

        /***********************************************************************

                Prohibit compress() from doing anything at all.

        ***********************************************************************/

        override IBuffer compress ()
        {
                return this;
        }               

        /***********************************************************************

                Prohibit clear() from doing anything at all.

        ***********************************************************************/

        override IBuffer clear ()
        {
                return this;
        }               

        /***********************************************************************
        
                Prohibit the setting of another IConduit

        ***********************************************************************/

        override void setConduit (IConduit conduit)
        {
                assert (0);
        }
}
