/*******************************************************************************

        @file DeviceConduit.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version; May 2005      

        @author         Kris 

*******************************************************************************/

module mango.io.DeviceConduit;

private import  mango.sys.OS,
                mango.sys.System;
                
public  import  mango.io.Buffer,
                mango.io.FileStyle;

private import  mango.io.Conduit,
                mango.io.Exception,
                mango.io.ConduitStyle;

/*******************************************************************************

        Bring in native Windows functions

*******************************************************************************/

version (Win32)
{
        private extern (Windows) 
        {
                HANDLE GetStdHandle (DWORD);
        }
}


/*******************************************************************************

        Implements a means of reading and writing a file device. Conduits
        are the primary means of accessing external data, and are usually 
        routed through a Buffer. 
        
*******************************************************************************/

class DeviceConduit : Conduit
{
        // expose conduit.copy() methods also
        alias Conduit.copy  copy;
        alias Conduit.read  read;
        alias Conduit.write write;

        /***********************************************************************
        
                Construct a conduit with the given style and seek abilities. 
                Conduits are either seekable or non-seekable.

        ***********************************************************************/

        this (ConduitStyle style, bool seekable)
        {
                super (style, seekable);
        }

        /***********************************************************************

                Create a FileConduit on the provided FileDevice. This is 
                strictly for adapting existing devices such as Stdout and
                friends.
        
        ***********************************************************************/

        this (FileDevice device)
        {
                // say we're not seekable
                super (device, false);
                
                // open the file
                reopen (device);

                // bump lock count
                acquire();
        }    

        /***********************************************************************
        
                Callback to close the file. This is invoked from the Resource
                base-class when the resource is being closed.

        ***********************************************************************/

        protected override void closure ()
        {       
                super.closure ();
                _close ();
        }    
                   
        /***********************************************************************

                Create a Buffer of the default FileConduit size, and
                associate it with this conduit

        ***********************************************************************/

        override IBuffer createBuffer()
        {
                const int DefaultFileBufferSize = 1024 * 16; 

                IBuffer buffer = new Buffer (DefaultFileBufferSize);

                buffer.setConduit (this);
                return buffer;
        } 

        /***********************************************************************

                Return the name of this device

        ***********************************************************************/

        protected char[] getName()
        {
                return "<device>";
        }


        /***********************************************************************

                Windows-specific code
        
        ***********************************************************************/

        version (Win32)
        {
                protected HANDLE handle;

                /***************************************************************

                        Throw an IOException noting the last error

                ***************************************************************/

                protected void error ()
                {
                        throw new IOException (getName() ~ ": " ~ System.error);
                }

                /***************************************************************

                        Gain access to the standard IO handles (console etc).

                ***************************************************************/

                protected void reopen (FileDevice device)
                {
                        handle = cast(HANDLE) device.id;
                }

                /***************************************************************

                        Close the underlying file

                ***************************************************************/

                protected void _close ()
                {
                        if (! CloseHandle (handle))
                              error ();
                }

                /***************************************************************

                        Read a chunk of bytes from the file into the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                override int read (void[] dst)
                {
                        DWORD read;
                        void *p = dst;

                        if (! ReadFile (handle, p, dst.length, &read, null))
                              error ();

                        if (read == 0 && dst.length > 0)
                            return Eof;
                        return read;
                }

                /***************************************************************

                        Write a chunk of bytes to the file from the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                override int write (void[] src)
                {
                        DWORD written;

                        if (! WriteFile (handle, src, src.length, &written, null))
                              error ();

                        return written;
                }
        }


        /***********************************************************************

                 Unix-specific code.
        
        ***********************************************************************/

        version (Posix)
        {
                protected int handle = -1;

                /***************************************************************

                        Throw an IOException noting the last error

                ***************************************************************/

                protected void error ()
                {
                        throw new IOException (getName() ~ ": " ~
                                               System.error);
                }

                /***************************************************************

                        Gain access to the standard IO handles (console etc).

                ***************************************************************/

                protected void reopen (FileDevice device)
                {
                        handle = device.id;
                }

                /***************************************************************

                        Close the underlying file

                ***************************************************************/

                protected void _close ()
                {
                        if (posix.close (handle) == -1)
                            error ();
                }

                /***************************************************************

                        Read a chunk of bytes from the file into the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                override int read (void[] dst)
                {
                        int read = posix.read (handle, dst, dst.length);
                        if (read == -1)
                            error ();
                        else
                           if (read == 0 && dst.length > 0)
                               return Eof;
                        return read;
                }

                /***************************************************************

                        Write a chunk of bytes to the file from the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                override int write (void[] src)
                {
                        int written = posix.write (handle, src, src.length);
                        if (written == -1)
                            error ();
                        return written;
                }
        }
}


/*******************************************************************************

        Class used to wrap an existing file-oriented handle, such as Stdout
        and its cohorts.

*******************************************************************************/

class FileDevice : ConduitStyle
{
        private uint id;

        this (uint id, Access access)
        {
                super (access);
                this.id = id;
        }
}


/*******************************************************************************

        Conduit for specifically handling the console devices. This takes
        care of certain implementation details on the Win32 platform.
        
*******************************************************************************/

class ConsoleConduit : DeviceConduit
{
        // expose conduit.copy() methods also 
        alias DeviceConduit.copy  copy;
        alias DeviceConduit.read  read;
        alias DeviceConduit.write write;

        /***********************************************************************

                Create a FileConduit on the provided FileDevice. This is 
                strictly for adapting existing devices such as Stdout and
                friends.
        
        ***********************************************************************/

        package this (FileDevice device)
        {
                super (device);
        }    

        /***********************************************************************

                Return the name of this device

        ***********************************************************************/

        protected char[] getName()
        {
                return "console device";
        }

        /***********************************************************************

                Windows-specific code
        
        ***********************************************************************/

        version(Win32)
        {
                /***************************************************************

                        Gain access to the standard IO handles (console etc).

                ***************************************************************/

                protected override void reopen (FileDevice device)
                {
                        static const DWORD[] id = [
                                                  cast(DWORD) -10, 
                                                  cast(DWORD) -11, 
                                                  cast(DWORD) -12
                                                  ];
                        static const char[][] f = [
                                                  "CONIN$\0", 
                                                  "CONOUT$\0", 
                                                  "CONOUT$\0"
                                                  ];

                        assert (device.id < 3);
                        handle = GetStdHandle (id[device.id]);
                        if (! handle)
                              handle = CreateFileA (f[device.id], 
                                       GENERIC_READ | GENERIC_WRITE,  
                                       FILE_SHARE_READ | FILE_SHARE_WRITE, 
                                       null, OPEN_EXISTING, 0, null);
                        if (! handle)
                              error ();
                }

                /***************************************************************

                        Write a chunk of bytes to the file from the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                override int write (void[] src)
                {
                        void* p = src.ptr;
                        DWORD len = src.length;
                        const int Max = 0x7fff;

                        // Win32 console has problems with content over
                        // 32KB in length, so handle it specifically
                        do {
                           DWORD i = len;
                           if (i > Max)
                               i = Max;
                           if (! WriteFile (handle, p, i, &i, null))
                                 error ();
                           len -= i;
                           p += i;
                           } while (len);

                        return src.length;
                }
        }
}


