/*******************************************************************************

        @file FileConduit.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version; March 2004      

        @author         Kris 
                        John Reimer
                        Anders F Bjorklund (Darwin patches)
                        Chris Sauls (Win95 file support)

*******************************************************************************/

module mango.io.FileConduit;

public  import  mango.io.FilePath,
                mango.io.FileProxy,
                mango.io.FileStyle;

private import  mango.sys.OS;

private import  mango.io.Buffer,
                mango.io.Conduit,
                mango.io.ConduitStyle,
                mango.io.DeviceConduit;

version (Win32)
        {
        private extern (Windows) 
                {
                BOOL   SetEndOfFile       (HANDLE);
                BOOL   UnmapViewOfFile    (LPCVOID);
                BOOL   FlushViewOfFile    (LPCVOID, DWORD);
                LPVOID MapViewOfFile      (HANDLE, DWORD, DWORD, DWORD, DWORD);
                HANDLE CreateFileMappingA (HANDLE, LPSECURITY_ATTRIBUTES, DWORD, DWORD, DWORD, LPCTSTR);
                }
        }
     else
        private extern (C) int ftruncate (int, int);


/*******************************************************************************

        Implements a means of reading and writing a generic file. Conduits
        are the primary means of accessing external data, and are usually 
        routed through a Buffer. File conduit extends the generic conduit
        by providing file-specific methods to set the file size, seek to a
        specific file position, and so on. Also provided is a class for
        creating a memory-mapped Buffer upon a file.     
        
        Serial input and output is straightforward. In this example we
        copy a file directly to the console:

        @code
        // open a file for reading
        FileConduit from = new FileConduit ("test.txt");

        // stream directly to console
        Stdout.conduit.copy (from);
        @endcode

        And here we copy one file to another:

        @code
        // open a file for reading
        FileConduit from = new FileConduit ("test.txt");

        // open another for writing
        FileConduit to = new FileConduit ("copy.txt", FileStyle.WriteTruncate);

        // copy file
        to.copy (from);
        @endcode
        
        FileConduit can just as easily handle random IO. Here we see how
        a Reader and Writer are used to perform simple input and output:

        @code
        // open a file for reading
        FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);

        // construct (binary) reader & writer upon this conduit
        Reader read = new Reader (fc);
        Writer write = new Writer (fc);

        int x=10, y=20;

        // write some data, and flush output since IO is buffered
        write (x) (y) ();

        // rewind to file start
        fc.seek (0);

        // read data back again, but swap destinations
        read (y) (x);

        assert (y==10);
        assert (x==20);

        fc.close();
        @endcode

        FileConduits can also be used directly, without Readers, Writers, or
        Buffers. To load a file directly into local-memory one might do this:

        @code
        // open file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create an array to house the entire file
        char[] content = new char[fc.length];

        // read the file content. Return value is the number of bytes read
        int bytesRead = fc.read (content);
        @endcode

        Conversely, one may write directly to a FileConduit, like so:

        @code
        // open file for writing
        FileConduit to = new FileConduit ("text.txt", FileStyle.WriteTruncate);

        // write an array of content to it
        int bytesWritten = fc.write (content);
        @endcode


        See File, FilePath, FileProxy, FileConst, FileScan, and FileSystem for 
        additional functionality related to file manipulation. 
        
        Doxygen has a hard time with D version() statements, so part of this 
        class is documented within FileConduit::VersionWin32 instead.

        Compile with -version=Win32SansUnicode to enable Win95 & Win32s file 
        support.
        
*******************************************************************************/

class FileConduit : DeviceConduit, ISeekable
{
        // the file we're working with 
        private FilePath path;

        // expose deviceconduit.copy() methods also 
        alias DeviceConduit.copy      copy;
        alias DeviceConduit.read      read;
        alias DeviceConduit.write     write;

        /***********************************************************************
        
                Create a FileConduit with the provided path and style.

        ***********************************************************************/

        this (char[] name, FileStyle style = FileStyle.ReadExisting)
        {
                this (new FilePath(name), style);
        }

        /***********************************************************************
        
                Create a FileConduit from the provided proxy and style.

        ***********************************************************************/

        this (FileProxy proxy, FileStyle style = FileStyle.ReadExisting)
        {
                this (proxy.getPath(), style);
        }

        /***********************************************************************
        
                Create a FileConduit with the provided path and style.

        ***********************************************************************/

        this (FilePath path, FileStyle style = FileStyle.ReadExisting)
        {
                // say we're seekable
                super (style, true);
                
                // remember who we are
                this.path = path;

                // open the file
                open (style);

                // bump lock count
                acquire();
        }    

        /***********************************************************************
        
                Return the FilePath used by this file.

        ***********************************************************************/

        FilePath getPath ()
        {
                return path;
        }               

        /***********************************************************************
        
                Move the file position to the given offset (from the file 
                start) and return the adjusted postion.

        ***********************************************************************/

        long seek (long offset)
        {
                return seek (offset, SeekAnchor.Begin);
        }               

        /***********************************************************************
                
                Return the current file position.
                
        ***********************************************************************/

        long getPosition ()
        {
                return seek (0, SeekAnchor.Current);
        }               

        /***********************************************************************
        
                Return the total length of this file.

        ***********************************************************************/

        long length ()
        {
                long    pos,    
                        ret;
                        
                pos = getPosition ();
                ret = seek (0, SeekAnchor.End);
                seek (pos);
                return ret;
        }               

        /***********************************************************************

                Transfer the content of another file to this one. Returns a
                reference to this class on success, or throws an IOException 
                upon failure.
        
        ***********************************************************************/

        FileConduit copy (FilePath source)
        {
                auto FileConduit fc = new FileConduit (source);
                super.copy (fc);
                return this;
        }               

        /***********************************************************************
        
                Return the name used by this file.

        ***********************************************************************/

        protected override char[] getName ()
        {
                return path.toString;
        }               


        /***********************************************************************

                Windows-specific code
        
        ***********************************************************************/

        version(Win32)
        {
                private bool appending;

                /***************************************************************

                        Open a file with the provided style.

                ***************************************************************/

                protected void open (FileStyle style)
                {
                        DWORD   attr,
                                share,
                                access,
                                create;

                        alias DWORD[] Flags;

                        static const Flags Access =  
                                        [
                                        0,                      // invalid
                                        GENERIC_READ,
                                        GENERIC_WRITE,
                                        GENERIC_READ | GENERIC_WRITE,
                                        ];
                                                
                        static const Flags Create =  
                                        [
                                        OPEN_EXISTING,          // must exist
                                        CREATE_ALWAYS,          // create always
                                        TRUNCATE_EXISTING,      // must exist
                                        CREATE_ALWAYS,          // (for appending)
                                        ];
                                                
                        static const Flags Share =   
                                        [
                                        FILE_SHARE_READ,
                                        FILE_SHARE_WRITE,
                                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                                        ];
                                                
                        static const Flags Attr =   
                                        [
                                        0,
                                        FILE_FLAG_RANDOM_ACCESS,
                                        FILE_FLAG_SEQUENTIAL_SCAN,
                                        0,
                                        FILE_FLAG_WRITE_THROUGH,
                                        ];

                        attr = Attr[style.cache];
                        share = Share[style.share];
                        create = Create[style.open];
                        access = Access[style.access];

                        version (Win32SansUnicode)
                                 handle = CreateFileA (path.toUtf8, access, share, 
                                                       null, create, 
                                                       attr | FILE_ATTRIBUTE_NORMAL,
                                                       cast(HANDLE) null);
                             else
                                handle = CreateFileW (path.toUtf16, access, share, 
                                                      null, create, 
                                                      attr | FILE_ATTRIBUTE_NORMAL,
                                                      cast(HANDLE) null);

                        if (handle == INVALID_HANDLE_VALUE)
                            error ();

                        // move to end of file?
                        if (style.open == style.Open.Append)
                            appending = true;
                }
                
                /***************************************************************

                        Write a chunk of bytes to the file from the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                int write (void[] src)
                {
                        DWORD written;

                        // try to emulate the Unix O_APPEND mode
                        if (appending)
                            SetFilePointer (handle, 0, null, SeekAnchor.End);
                        
                        return super.write (src);
                }
            
                /***************************************************************

                        Set the file size to be that of the current seek 
                        position. The file must be writable for this to
                        succeed.

                ***************************************************************/

                void truncate ()
                {
                        // must have Generic_Write access
                        if (! SetEndOfFile (handle))
                              error ();                            
                }               

                /***************************************************************

                        Set the file seek position to the specified offset
                        from the given anchor. 

                ***************************************************************/

                long seek (long offset, SeekAnchor anchor)
                {
                        LONG high = (offset >> 32);
                        long result = SetFilePointer (handle, cast(LONG) offset, 
                                                      &high, anchor);

                        if (result == -1 && 
                            GetLastError() != ERROR_SUCCESS)
                            error ();

                        return result + (cast(long) high << 32);
                }               
        }


        /***********************************************************************

                 Unix-specific code. Note that some methods are 32bit only
        
        ***********************************************************************/

        version (Posix)
        {
                /***************************************************************

                        Open a file with the provided style.

                ***************************************************************/

                protected void open (FileStyle style)
                {
                        int     share, 
                                access;

                        alias int[] Flags;

                        static const Flags Access =  
                                        [
                                        0,              // invalid
                                        O_RDONLY,
                                        O_WRONLY,
                                        O_RDWR,
                                        ];
                                                
                        static const Flags Create =  
                                        [
                                        0,              // open existing
                                        O_CREAT,        // create always
                                        O_TRUNC,        // must exist
                                        O_APPEND | O_CREAT, 
                                        ];

                        // this is not the same as Windows sharing,
                        // but it's perhaps a reasonable approximation
                        static const Flags Share =   
                                        [
                                        0640,           // read access
                                        0620,           // write access
                                        0660,           // read & write
                                        ];
                                                
                        share = Share[style.share];
                        access = Access[style.access] | Create[style.open];

                        handle = posix.open (path.toUtf8, access, share);
                        if (handle == -1)
                            error ();
                }

                /***************************************************************

                        32bit only ...

                        Set the file size to be that of the current seek 
                        position. The file must be writable for this to
                        succeed.

                ***************************************************************/

                void truncate ()
                {
                        // set filesize to be current seek-position
                        if (ftruncate (handle, getPosition()) == -1)
                            error ();
                }               

                /***************************************************************

                        32bit only ...

                        Set the file seek position to the specified offset
                        from the given anchor. 

                ***************************************************************/

                long seek (long offset, SeekAnchor anchor)
                {
                        int result = posix.lseek (handle, offset, anchor);
                        if (result == -1)
                            error ();
                        return result;
                }               
        }
}



/*******************************************************************************

        Class to handle memory-mapped files

*******************************************************************************/
version (Mapped)
{
class MappedFile : MappedBuffer
{
        version (Win32)
        {
                private FileConduit     host;
                private void*           base;
                private HANDLE          mmFile;

                /***************************************************************

                        Construct a MappedBuffer upon the given FileConduit. 
                        One should set the file size using seek() & truncate() 
                        to setup the available working space.

                ***************************************************************/

                this (FileConduit host)
                {
                        this.host = host;

                        // can only do 32bit mapping on 32bit platform
                        long size = host.length;
                        ConduitStyle style = host.getStyle;

                        DWORD flags = PAGE_READWRITE;
                        if (style.access == ConduitStyle.Access.Read)
                            flags = PAGE_READONLY;

                        mmFile = CreateFileMappingA (host.handle, null, flags, 0, 0, null);
                        if (mmFile is null)
                            host.error ();

                        flags = FILE_MAP_WRITE;
                        if (style.access == ConduitStyle.Access.Read)
                            flags = FILE_MAP_READ;

                        base = MapViewOfFile (mmFile, flags, 0, 0, 0);
                        if (base is null)
                            host.error ();

                        void[] mem = base[0..size];
                        setValidContent (mem);
                }

                /***************************************************************

                        Ensure this is closed when GC'd

                ***************************************************************/

                ~this ()
                {
                        close ();        
                }

                /***************************************************************

                        Close this mapped buffer

                ***************************************************************/

                void close ()
                {
                        if (base)
                            UnmapViewOfFile (base);

                        if (mmFile)
                            CloseHandle (mmFile);       

                        mmFile = null;
                        base = null;
                }

                /***************************************************************

                        Flush dirty content out to the drive. This
                        fails with error 33 if the file content is
                        virgin. Opening a file for ReadWriteExists
                        followed by a flush() will cause this.

                ***************************************************************/

                void flush ()
                {
                        // flush all dirty pages
                        if (! FlushViewOfFile (base, 0))
                              host.error ();
                }
        }


        version (Posix)
        {
                // not yet supported
                private this (){};
        }
}
}
