/*******************************************************************************

        @file SocketConduit.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
			Jan 1st 2005 - Added RedShodan patch for timeout query

        @author         Kris

*******************************************************************************/

module mango.io.SocketConduit;

version = UseFreeList;

public  import  mango.io.Socket;

private import  mango.io.Buffer,
                mango.io.Exception,
                mango.io.ConduitStyle;

private import  mango.io.model.IConduit;

/*******************************************************************************

        A wrapper around the bare Socket to implement the IConduit abstraction
        and add socket-specific functionality. SocketConduit data-transfer is 
        typically performed in conjunction with an IBuffer instance, but can 
        be handled directly using raw arrays if preferred. See FileConduit for 
        examples of both approaches.

*******************************************************************************/

class SocketConduit : Socket, IConduit, ISocketReader
{       
        private timeval         tv;
        private SocketSet       ss;
        private IConduitSource  input;
        private IConduitSink    output;
	private bool		timeout;

        version (UseFreeList)
        {
                /***************************************************************

                        Instance variables for free-list support

                ***************************************************************/

                private SocketConduit           next;   
                private bool                    fromList;
                private static SocketConduit    freelist;

                /***************************************************************

                        Allocate a SocketConduit from a list rather than 
                        creating a new one

                ***************************************************************/

                private static synchronized SocketConduit allocate (socket_t sock)
                {       
                        SocketConduit s;

                        if (freelist)
                           {
                           s = freelist;
                           freelist = s.next;
                           s.set (sock);
                           }
                        else
                           {
                           s = new SocketConduit (sock);
                           s.fromList = true;
                           }
                        return s;
                }

                /***************************************************************

                        Return this SocketConduit to the free-list

                ***************************************************************/

                private static synchronized void deallocate (SocketConduit s)
                {
                        // socket handle is no longer valid
                        s.reset ();
                        s.next = freelist;
                        freelist = s;
                }

                /***************************************************************

                        Override closure() to deallocate this SocketConduit 
                        when it has been closed. Note that one should *not* 
                        delete a SocketConduit when FreeList is enabled ...

                ***************************************************************/

                protected override void closure ()
                {       
                        input.unbind ();
                        output.unbind ();
                        fixup ();

                        // be a nice client, and tell the server?
                        //super.shutdown();

                        // do this first cos' we're gonna' reset the
                        // socket handle during deallocate()
                        super.closure ();

                        // deallocate if this came from the free-list,
                        // otherwise just wait for the GC to handle it
                        if (fromList)
                            deallocate (this);
                }
        }
        else
        {
                /***************************************************************
        
                        Method to close the filters. This is invoked from the 
                        Resource base-class when the resource is being closed.
                        You should ensure that a subclass invokes this as part
                        of its closure mechanics.

                ***************************************************************/

                protected override void closure ()
                {       
                        input.unbind ();
                        output.unbind ();
                }    
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void fixup ()
        {
                input = this;
                output = this;
        }
                        
        /***********************************************************************
        
                Construct this SocketConduit with the given socket handle;
                this is for FreeList and ServerSocket support.

        ***********************************************************************/

        static SocketConduit create (socket_t handle)
        {
                version (UseFreeList)
                        {
                        // allocate one from the free-list
                        return allocate (handle);
                        }
                     else
                        {
                        // create a new SocketConduit instance
                        return new SocketConduit (handle);
                        }
        }

        /***********************************************************************
        
                Create a streaming Internet Socket

        ***********************************************************************/

        this ()
        {
                super (AddressFamily.INET, Type.STREAM, Protocol.IP);                
                fixup ();
        }

        /***********************************************************************
        
                Construct this SocketConduit with the given socket handle;
                this is for FreeList and ServerSocket support.

        ***********************************************************************/

        private this (socket_t handle)
        {
                super (handle);                
                fixup ();
        }

        /***********************************************************************
        
                Create a Buffer of a conduit-specific size. I think 8K bytes
                is the common standard for socket buffering?

        ***********************************************************************/

        IBuffer createBuffer ()
        {
                Buffer buffer = new Buffer (1024 * 8);
                buffer.setConduit (this);
                return buffer;
        }

        /***********************************************************************
        
                Set the read timeout to the specified microseconds. Set a 
                value of zero to disable timeout support.

        ***********************************************************************/

        void setTimeout (uint us)
        {
                tv.tv_sec = 0;
                tv.tv_usec = us;
        }
                         
        /***********************************************************************
		
		Did the last operation result in a timeout? Note that this
		assumes there is no thread contention on this object.        

        ***********************************************************************/

        bool hadTimeout ()
        {
		return timeout;
        }
                         
        /***********************************************************************
        
                Please refer to IConduit.attach for details

        ***********************************************************************/

        void attach (IConduitSource input)
        {
                input.bind (this.input);
                this.input = input;
        }

        /***********************************************************************
                
                Please refer to IConduit.attach for details

        ***********************************************************************/

        void attach (IConduitSink output)
        {
                output.bind (this.output);
                this.output = output;
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected void bind (IConduitSink sink)
        {
        }
                              
        /***********************************************************************
        
        ***********************************************************************/

        protected void bind (IConduitSource source)
        {
        }                            
                              
        /***********************************************************************
        
        ***********************************************************************/

        protected void unbind ()
        {
        }

        /***********************************************************************
        
                Callback routine to read content from the socket. Note 
                that the operation may timeout if method setTimeout()
                has been invoked with a non-zero value.         

                Returns the number of bytes read from the socket, or
                IConduit.Eof where there's no more content available

		Note that a timeout is equivalent to Eof. Isolating
		a timeout condition can be achieved via hadTimeout() 

		Note also that a zero return value is not legitimate; 
		such a value indicates Eof

        ***********************************************************************/

        int read (void[] src)
        {
                // ensure just one read at a time 
                synchronized (_lock)
                {
		// reset timeout; we assume there's no thread contention
		timeout = false;

                // did user disable timeout checks?
                if (tv.tv_usec)
                   {
                   // nope: ensure we have a SocketSet
                   if (ss is null)
                       ss = new SocketSet (1);

                   ss.reset ();
                   ss.add (this);

                   // wait until data is available, or a timeout occurs
                   int i = select (ss, null, null, &tv);
		   if (i <= 0)
                      {
		      if (i == 0)
			  timeout = true;
                      return Eof;
                      }
                   }

                int count = receive (src);
                if (count <= 0)
                    count = Eof;
                return count;
                }
        }

        /***********************************************************************
        
                Callback routine to write the provided content to the 
                socket. This will stall until the socket responds in
                some manner. Returns the number of bytes sent to the
                output, or IConduit.Eof if the socket cannot write.

        ***********************************************************************/

        int write (void[] src)
        {
                int count = send (src);
                if (count <= 0)
                    count = Eof;
                return count;
        }

        /***********************************************************************
        
                Read from conduit into a target buffer. Note that this 
                uses SocketSet to handle timeouts, such that the socket
                does not stall forever.

        ***********************************************************************/

        int read (IBuffer target)
        in {
           assert (target);
           }
        body
        {
                return target.write (&input.read);
        }

        /***********************************************************************
        
                Write to conduit from a source buffer.

        ***********************************************************************/

        int write (IBuffer source)
        in {
           assert (source);
           }
        body
        {
                int count = source.read (&output.write);

                // if we didn't write everything, move remaining content
                // to front of buffer for a subsequent write
                source.compress();
                return count;
        }

        /***********************************************************************
        
                This conduit is not seekable

        ***********************************************************************/

        bool isSeekable ()
        {
                return false;
        }               

        /***********************************************************************
        
                This conduit is readable

        ***********************************************************************/

        bool isReadable ()
        {
                return true;
        }               

        /***********************************************************************
        
                This conduit is writable

        ***********************************************************************/

        bool isWritable ()
        {
                return true;
        }               

        /***********************************************************************

                Transfer the content of this conduit to another one. 
                Returns true if all content was successfully copied.
        
        ***********************************************************************/

        IConduit copy (IConduit source)
        in {
           assert (source);
           }
        body
        {
                IBuffer buffer = createBuffer ();

                while (source.read (buffer) != Eof)
                       if (write (buffer) == Eof)
                           throw new IOException ("Eof while copying to Socket");

                // flush any remains into the target
                return flush (buffer);
        }               

        /***********************************************************************
        
                Flush buffer content out to this conduit.

                Returns true if all content is flushed; false if writing 
                results in an Eof condition.
                        
        ***********************************************************************/

        IConduit flush (IBuffer source)
        {
                while (source.readable())
                       if (write (source) == Eof)
                           throw new IOException ("Eof while flushing Socket");
                return this;
        }               

        /***********************************************************************
        
                Return the style used to open the conduit

        ***********************************************************************/

        ConduitStyle getStyle ()
        {
                return ConduitStyle.ReadWrite;
        }    
}



