/*******************************************************************************

        @file TextLayout.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris
                        Anders F Bjorklund (Darwin patches)


*******************************************************************************/

module mango.io.TextLayout;

private import  mango.io.Exception;

/*******************************************************************************

        Arranges text strings in order, using indices to specify where
        each particular argument should be positioned within the text. 
        This is handy for collating I18N components.

        @code
        // create a TextLayout instance
        TextLayout tl = new TextLayout (256);

        // write text to Stdout
        char[] arg1 = "one";
        char[] arg2 = "two";
        Stdout (tl.format ("%2 %1", arg1, arg2));

        // or access the content, post processing ...
        char[] text = tl.content;
        @endcode

        The index numbers range from one through nine. Note also that 
        this class is stateful, and is therefore not shareable across 
        multiple threads.
        
*******************************************************************************/

class TextLayout (Type)
{       
        alias format opCall;

        private int     count;
        private Type[]  output;
                       
        /***********************************************************************
        
                Set the maximum buffer length

        ***********************************************************************/

        this (int limit)
        {
                output = new Type [limit];
        }

        /***********************************************************************
        
                Address formatted output as a char[]

        ***********************************************************************/

        Type[] content ()
        {
                return output [0..count];
        }

        /***********************************************************************
                
                layout a set of arguments

        ***********************************************************************/

        Type[] format (Type[][] layout ...)
        {
                int     pos,
                        args;
                bool    state;
                        
                args = layout.length - 1;
                foreach (Type c; layout[0])
                        {
                        if (state)
                           {
                           state = false;
                           if (c >= '1' || c <= '9')
                              {
                              uint index = c - '0';
                              if (index <= args)
                                 {
                                 Type[] x = layout[index];

                                 int limit = pos + x.length;
                                 if (limit < output.length)
                                    {
                                    output[pos..limit] = x;
                                    pos = limit;
                                    continue;
                                    } 
                                 else
                                    throw new Exception ("TextLayout limit exceeded");
                                 }
                              else
                                 throw new Exception ("invalid layout argument");
                              }
                           }
                        else
                           if (c == '%')
                              {
                              state = true;
                              continue;
                              }
                
                        if (pos < output.length)
                           {
                           output[pos] = c;
                           ++pos;
                           }
                        else     
                           throw new Exception ("layout buffer too small");
                        }

                count = pos;
                return content;
        }
}

alias TextLayout!(char)  LayoutChar;
alias TextLayout!(wchar) LayoutWChar;
alias TextLayout!(dchar) LayoutDChar;


