/*******************************************************************************

        @file TextReader.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.TextReader;

private import  mango.sys.Type;

private import  mango.io.Token,
                mango.io.Reader,
                mango.io.Tokenizer,
                mango.io.Exception;

/*******************************************************************************

        Grok readable input from a stream. All input is tokenized from the
        associated buffer, and converted as necessary into the destination 
        location. A SpaceTokenizer is used by default, but you can choose
        an alternative (such as a comma-delimiting tokenizer).

*******************************************************************************/

class TextReader : Reader
{
        alias Reader.get get;

        // an individual token
        private Token token;

        // internal tokenizer instance
        private ITokenizer tokenizer;

        private final static char[] EOI = "unexpected end of input";

        /***********************************************************************
        
                Construct a TextReader on the provided buffer, using the
                specified Tokenizer instead of the default one.

        ***********************************************************************/

        this (IBuffer buffer, ITokenizer tokenizer = Tokenizers.line)
        {
                super (buffer);
                this.token = new Token;
                this.tokenizer = tokenizer;
        }
        
        /***********************************************************************
                
                Construct a TextReader upon the buffer associated with the
                given conduit.

        ***********************************************************************/

        this (IConduit conduit, ITokenizer tokenizer = Tokenizers.line)
        {
                this (conduit.createBuffer, tokenizer);
        }

        /***********************************************************************
        
        ***********************************************************************/

        override uint read (void* src, uint bytes, uint type)
        {
                int length = bytes;

                // get width of elements (note: does not work for bit[])
                int width = Type.widths[type];

                // for all bytes in source ...
                while (bytes)
                      {
                      switch (type)
                             {
                             case Type.Bool:
                                  *cast(bool*) src = (next.toString() == "true");
                                  break;

                             case Type.Byte:
                             case Type.UByte:
                                  *cast(ubyte*) src = next.toInt();
                                  break;

                             case Type.Short:
                             case Type.UShort:
                                  *cast(ushort*) src = next.toInt();
                                  break;

                             case Type.Int:
                             case Type.UInt:
                                  *cast(uint*) src = next.toInt();
                                  break;

                             case Type.Long:
                             case Type.ULong:
                                  *cast(ulong*) src = next.toLong();
                                  break;

                             case Type.Float:
                                  *cast(float*) src = next.toReal();
                                  break;

                             case Type.Double:
                                  *cast(double*) src = next.toReal();
                                  break;

                             case Type.Real:
                                  *cast(real*) src = next.toReal();
                                  break;

                             default:
                                  error ("unknown type handed to TextReader.read()");
                             }

                      // bump counters and loop around for next instance
                      bytes -= width;
                      src += width;
                      }

                return length;
        }

        /***********************************************************************
        
        ***********************************************************************/

        override IReader get (inout char[] x, uint elements = uint.max) 
        { 
                // dup the input string, to avoid surprises
                x = nextToken.toString (false);
                return this;
        }

        /***********************************************************************
        
                @todo - Tokenizer needs to handle wchar[] before this will
                        operate correctly

        ***********************************************************************/

        override IReader get (inout wchar[] x, uint elements = uint.max) 
        { 
                assert (0);
                return this;
        }

        /***********************************************************************
        
                @todo - Tokenizer needs to handle dchar[] before this will
                        operate correctly

        ***********************************************************************/

        override IReader get (inout dchar[] x, uint elements = uint.max) 
        { 
                assert (0);
                return this;
        }

        /***********************************************************************
        
                Throw an exception if the input requested is not available.
                Reader instances expect the input to be available, so you 
                might think of them as being applied to a fixed-format 
                file. Tokenizers, on the other hand, are more flexible in
                that they return true until no more tokens are available.
                      
                Readers take care of assignment to user-space variables, 
                whereas Tokenizers just make the raw content available.
                
        ***********************************************************************/

        private static final void error (char[] msg = EOI)
        {
                throw new TokenException (msg);
        }

        /***********************************************************************
        
                Internal method to capture the next token.

        ***********************************************************************/

        private final Token nextToken ()
        {
                if (! tokenizer.next (buffer, token))
                      error (); 
                return token;
        }

        /***********************************************************************
        
                Internal method to isolate the next token, and check its
                length. This is used when converting to non-char types, 
                such as integers ~ an empty token is illegal, since we
                don't support default values.

                @todo: should the empty-token rule be relaxed?

        ***********************************************************************/

        private final Token next ()
        {
                nextToken();
                if (token.getLength == 0)
                    error ();
                return token;    
        }
}
