/*******************************************************************************

        @file Utf.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, October 2004      
        @author         Kris


*******************************************************************************/

module mango.io.Utf;

private import mango.io.Exception;

/*******************************************************************************

        Fast UTF-8 to Unicode transcoder. These are really sensitive to
        small changes on 32bit x86 devices, because the register set of
        those devices is so small. Beware of subtle changes which might
        extend the execution-period by as much as 200% ...

        These routines were tuned on an Intel P3; other devices may work
        more efficiently with a slightly different approach, though this
        is likely to be reasonably optimal on AMD x86 CPUs also. These
        algorithms could benefit significantly from those extra AMD64 
        registers.

        Note that foreach can produce noticeable more efficient code than 
        equivalent for() loops, with either indices or pointers. The 0.98
        compiler version exhibited some rather erratic behavior over the
        course of testing: in particular, elapsed time of method execution
        is noticeably dependent upon its physical location within the file
        (or, more specifically, the enclosing class). Yes, it sure sounds 
        crazy that if you switch the order of encode() with decode() that 
        they will consistently execute slower than as currently arranged.        

        Finally, please note that these are between 5 and 30 times faster 
        than equivalent functions in the std.utf Phobos module (dependent
        upon the mix of char values). Those functions (strangely) often
        allocate memory on a character basis, so will become significantly 
        slower where there's heap-contention by multiple threads.
        
*******************************************************************************/

class Utf
{
        /***********************************************************************

        ***********************************************************************/

        private static final void error (char[] msg)
        {
                throw new IOException (msg);
        }

        /***********************************************************************

                Encode Utf8 up to a maximum of 3 bytes long (four, five & 
                six byte variations are not supported). Throws an exception
                where the input wchar is greater than 0xd800.

                If the output is provided off the stack, it should be large 
                enough to encompass the entire utf8 encoding. This option is 
                provided purely as an optimization for those cases where all 
                boundary conditions are explicitly checked for by the caller.

        ***********************************************************************/

        static final char[] toUtf8 (wchar[] input, char[] output = null)
        {
                if (! output.length)
                      output = new char [input.length * 2 + 3];

                char* pOut = output;
                char* pMax = pOut + output.length - 3;

                foreach (wchar b; input)
                        { 
                        if (pOut >= pMax)
                           {
                           int len = pOut - output.ptr;
                           output.length = len + len / 2;
                           pOut = output.ptr + len;
                           pMax = pOut + len - 4;
                           }

                        if (b < 0x80)
                            *pOut++ = b;
                        else
                           if (b < 0x0800)
                              {
                              pOut[0] = 0xc0 | ((b >> 6) & 0x3f);
                              pOut[1] = 0x80 | (b & 0x3f);
                              pOut += 2;
                              }
                           else
                              if (b < 0xd800)
                                 {
                                 pOut[0] = 0xe0 | ((b >> 12) & 0x3f);
                                 pOut[1] = 0x80 | ((b >> 6)  & 0x3f);
                                 pOut[2] = 0x80 | (b & 0x3f);
                                 pOut += 3;
                                 }
                              else
                                 error ("invalid wchar");
                        }
                
                return output [0..(pOut - output.ptr)];
        }


        /***********************************************************************

                Decode Utf8 produced by the above toUtf8() method. 
        
                If the output is provided off the stack, it should be large 
                enough to encompass the entire utf8 encoding. This option is 
                provided purely as an optimization for those cases where all 
                boundary conditions are explicitly checked for by the caller.

        ***********************************************************************/

        static final wchar[] toUtf16 (char[] input, wchar[] output = null)
        {
                int     produced;
                char*   pIn = input;
                char*   pMax = pIn + input.length;

                if (! output.length)
                      output = new wchar[input.length];

                foreach (inout wchar d; output)
                        {
                        wchar b = cast(wchar) *pIn;
                        if (b & 0x80)
                            if (b < 0xe0)
                               {
                               b &= 0x1f;
                               b = (b << 6) | (*++pIn & 0x3f);
                               }
                            else
                               if (b < 0xf0)
                                  {
                                  b &= 0x0f;
                                  b = (b << 6) | (pIn[1] & 0x3f);
                                  b = (b << 6) | (pIn[2] & 0x3f);
                                  pIn += 2;
                                  }

                        d = b;
                        ++produced;

                        if (++pIn >= pMax)
                            if (pIn > pMax)    
                                error ("invalid utf8");  
                            else
                               break;
                        }
                       
                if (pIn < pMax)
                    error ("utf8 overflow");
                return output [0..produced];
        }


        /***********************************************************************

                Encode Utf8 up to a maximum of 4 bytes long (five & six
                byte variations are not supported). Throws an exception
                where the input dchar is greater than 0x10ffff.

                If the output is provided off the stack, it should be large 
                enough to encompass the entire utf8 encoding. This option is 
                provided purely as an optimization for those cases where all 
                boundary conditions are explicitly checked for by the caller.

        ***********************************************************************/

        static final char[] toUtf8 (dchar[] input, char[] output = null)
        {
                if (! output.length)
                      output = new char [input.length * 2 + 4];

                char* pOut = output;
                char* pMax = pOut + output.length - 4;

                foreach (dchar b; input)
                        { 
                        if (pOut >= pMax)
                           {
                           int len = pOut - output.ptr;
                           output.length = len + len / 2;
                           pOut = output.ptr + len;
                           pMax = pOut + len - 4;
                           }

                        if (b < 0x80)
                            *pOut++ = b;
                        else
                           if (b < 0x0800)
                              {
                              pOut[0] = 0xc0 | ((b >> 6) & 0x3f);
                              pOut[1] = 0x80 | (b & 0x3f);
                              pOut += 2;
                              }
                           else
                              if (b < 0x80000)
                                 {
                                 pOut[0] = 0xe0 | ((b >> 12) & 0x3f);
                                 pOut[1] = 0x80 | ((b >> 6)  & 0x3f);
                                 pOut[2] = 0x80 | (b & 0x3f);
                                 pOut += 3;
                                 }
                              else
                                 if (b < 0x110000)
                                    {
                                    pOut[0] = 0xf0 | ((b >> 18) & 0x3f);
                                    pOut[1] = 0x80 | ((b >> 12) & 0x3f);
                                    pOut[2] = 0x80 | ((b >> 6)  & 0x3f);
                                    pOut[3] = 0x80 | (b & 0x3f);
                                    pOut += 4;
                                    }
                                 else
                                    error ("invalid dchar");
                        }
                
                return output [0..(pOut - output.ptr)];
        }


        /***********************************************************************

                Decode Utf8 produced by the above toUtf8() method. 
        
                If the output is provided off the stack, it should be large 
                enough to encompass the entire utf8 encoding. This option is 
                provided purely as an optimization for those cases where all 
                boundary conditions are explicitly checked for by the caller.

        ***********************************************************************/

        static final dchar[] toUtf32 (char[] input, dchar[] output = null)
        {
                int     produced;
                char*   pIn = input;
                char*   pMax = pIn + input.length;
                
                if (! output.length)
                      output = new dchar[input.length];

                foreach (inout dchar d; output)
                        {
                        dchar b = cast(dchar) *pIn;
                        if (b & 0x80)
                            if (b < 0xe0)
                               {
                               b &= 0x1f;
                               b = (b << 6) | (*++pIn & 0x3f);
                               }
                            else
                               if (b < 0xf0)
                                  {
                                  b &= 0x0f;
                                  b = (b << 6) | (pIn[1] & 0x3f);
                                  b = (b << 6) | (pIn[2] & 0x3f);
                                  pIn += 2;
                                  }
                               else
                                  {
                                  b &= 0x07;
                                  b = (b << 6) | (pIn[1] & 0x3f);
                                  b = (b << 6) | (pIn[2] & 0x3f);
                                  b = (b << 6) | (pIn[3] & 0x3f);
                                  pIn += 3;
                                  }

                        d = b;
                        ++produced;

                        if (++pIn >= pMax)
                            if (pIn > pMax)    
                                error ("invalid utf8");  
                            else
                               break;
                        }

                if (pIn < pMax)
                    error ("utf8 overflow");
                return output [0..produced];
        }
}
