/*******************************************************************************

        @file IBuffer.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.model.IBuffer;

private import mango.io.model.IConduit;


/*******************************************************************************

        The basic premise behind this IO package is as follows:

        1) the central concept is that of a buffer. The buffer acts
           as a queue (line) where items are removed from the front
           and new items are added to the back. Buffers are modeled 
           by this interface, and mango.io.Buffer exposes a concrete 
           implementation.

        2) buffers can be written to directly, but a Reader and/or
           Writer are typically used to read & write formatted data.
           These readers & writers are bound to a specific buffer;
           often the same buffer. It's also perfectly legitimate to 
           bind multiple writers to the same buffer; they will all
           behave serially as one would expect. The same applies to
           multiple readers on the same buffer. Readers and writers
           support two styles of IO: put/get, and the C++ style << 
           and >> operators. All such operations can be chained.

        3) Any class can be made compatable with the reader/writer
           framework by implementing the IReadable and/or IWritable 
           interfaces. Each of these specify just a single method.

        4) Buffers may also be tokenized. This is handy when one is
           dealing with text input, and/or the content suits a more
           fluid format than most typical readers & writers support.
           Tokens are mapped directly onto buffer content, so there
           is only minor overhead in using them. Tokens can be read
           and written by reader/writers also, using a more relaxed
           set of rules than those applied to integral IO.

        5) buffers are sometimes memory-only, in which case there
           is nothing left to do when a reader (or tokenizer) hits
           end of buffer conditions. Other buffers are themselves 
           bound to a Conduit. When this is the case, a reader will 
           eventually cause the buffer to reload via its associated 
           conduit. Previous buffer content will thus be lost. The
           same concept is applied to writers, whereby they flush 
           the content of a full buffer to a bound conduit before 
           continuing. 

        6) conduits provide virtualized access to external content,
           and represent things like files or Internet connections.
           They are just a different kind of stream. Conduits are
           modelled by mango.io.model.IConduit, and implemented via
           classes FileConduit and SocketConduit. Additional kinds
           of conduit are easy to construct: one either subclasses
           mango.io.Conduit, or implements mango.io.model.IConduit. A
           conduit reads and writes from/to a buffer in big chunks
           (typically the entire buffer).

*******************************************************************************/

interface IBuffer
{
        /***********************************************************************
                
                Create an instance of an IBuffer. Use this when you
                don't know anything about the concrete implementation,
                and have only the IBuffer interface available

        ***********************************************************************/

        IBuffer create ();

        /***********************************************************************
                
                Return the backing array

        ***********************************************************************/

        void[] getContent ();

        /***********************************************************************
        
                Return a char[] slice of the buffer up to the limit of
                valid content.

        ***********************************************************************/

        char[] toString ();

        /***********************************************************************
        
                Set the backing array with all content readable. Writing
                to this will either flush it to an associated conduit, or
                raise an Eof condition. Use IBuffer.clear() to reset the
                content (make it all writable).

        ***********************************************************************/

        IBuffer setValidContent (void[] data);

        /***********************************************************************
        
                Set the backing array with some content readable. Writing
                to this will either flush it to an associated conduit, or
                raise an Eof condition. Use IBuffer.clear() to reset the
                content (make it all writable).

        ***********************************************************************/

        IBuffer setContent (void[] data, uint readable);

        /***********************************************************************

                Write an array of data into this buffer, and flush to the
                conduit as necessary. Returns a chaining reference if all 
                data was written; throws an IOException indicating eof or 
                eob if not.

                This is often used in lieu of a Writer.

        ***********************************************************************/

        IBuffer append (void[] content);

        /***********************************************************************

                Read a chunk of data from the buffer, loading from the
                conduit as necessary. The requested number of bytes are
                loaded into the buffer, and marked as having been read 
                when the 'eat' parameter is set true. When 'eat' is set
                false, the read position is not adjusted.

                Returns the corresponding buffer slice when successful, 
                or null if there's not enough data available (Eof; Eob).

        ***********************************************************************/

        void[] get (uint size, bool eat = true);

        /***********************************************************************

                Exposes the raw data buffer at the current write position, 
                The delegate is provided with a void[] representing space
                available within the buffer at the current write position.

                The delegate should return the approriate number of bytes 
                if it writes valid content, or IConduit.Eof on error.

                Returns whatever the delegate returns.

        ***********************************************************************/

        int write (int delegate (void[]) writer);

        /***********************************************************************

                Exposes the raw data buffer at the current read position. The
                delegate is provided with a void[] representing the available
                data, and should return zero to leave the current read position
                intact. 
                
                If the delegate consumes data, it should return the number of 
                bytes consumed; or IConduit.Eof to indicate an error.

                Returns whatever the delegate returns.

        ***********************************************************************/

        int read (int delegate (void[]) reader);

        /***********************************************************************

                If we have some data left after an export, move it to 
                front-of-buffer and set position to be just after the 
                remains. This is for supporting certain conduits which 
                choose to write just the initial portion of a request.
                            
                Limit is set to the amount of data remaining. Position 
                is always reset to zero.

        ***********************************************************************/

        IBuffer compress ();

        /***********************************************************************
        
                Skip ahead by the specified number of bytes, streaming from 
                the associated conduit as necessary.
        
                Can also reverse the read position by 'size' bytes. This may
                be used to support lookahead-type operations.

                Returns true if successful, false otherwise.

        ***********************************************************************/

        bool skip (int size);

        /***********************************************************************
        
                flush the contents of this buffer to the related conduit.
                Throws an IOException on premature eof.

        ***********************************************************************/

        void flush ();

        /***********************************************************************
        
                Reset position and limit to zero.

        ***********************************************************************/

        IBuffer clear ();               

        /***********************************************************************
        
                return count of readable bytes remaining in buffer. This is 
                calculated simply as limit() - position()

        ***********************************************************************/

        uint readable ();               

        /***********************************************************************
        
                Return count of writable bytes available in buffer. This is 
                calculated simply as capacity() - limit()

        ***********************************************************************/

        uint writable ();

        /***********************************************************************
        
                returns the limit of readable content within this buffer

        ***********************************************************************/

        uint getLimit ();               

        /***********************************************************************
        
                returns the total capacity of this buffer

        ***********************************************************************/

        uint getCapacity ();               

        /***********************************************************************
        
                returns the current position within this buffer

        ***********************************************************************/

        uint getPosition ();               

        /***********************************************************************

                Overridable method to grow the buffer size when it becomes 
                full. Default is to not grow at all.

        ***********************************************************************/

        bool grow (uint size);

        /***********************************************************************
        
                Returns the conduit associated with this buffer. Returns 
                null if the buffer is purely memory based; that is, it's
                not backed by some external conduit.

                Buffers do not require a conduit to operate, but it can
                be convenient to associate one. For example, the IReader
                and IWriter classes use this to import/export content as
                necessary.

        ***********************************************************************/

        IConduit getConduit ();               

        /***********************************************************************
        
                Sets the external conduit associated with this buffer.

                Buffers do not require an external conduit to operate, but 
                it can be convenient to associate one. For example, methods
                read and write use it to import/export content as necessary.

        ***********************************************************************/

        void setConduit (IConduit conduit);
}

/*******************************************************************************

*******************************************************************************/

enum ConverterType 
{
        Char=0, WChar, DChar
};

