/*******************************************************************************

        @file IConduit.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.model.IConduit;

private import  mango.io.model.IBuffer;

private import  mango.io.model.IResource;

/*******************************************************************************

        Conduits provide virtualized access to external content, and 
        represent things like files or Internet connections. Conduits 
        are modelled by mango.io.model.IConduit, and implemented via 
        classes FileConduit and SocketConduit. 
        
        Additional kinds of conduit are easy to construct: one either 
        subclasses mango.io.Conduit, or implements mango.io.model.IConduit. 
        A conduit typically reads and writes from/to an IBuffer in large 
        chunks, typically the entire buffer. Alternatively, one can invoke 
        read(dst[]) and/or write(src[]) directly.

*******************************************************************************/

interface IConduit : IResource, IConduitSource, IConduitSink
{
        /***********************************************************************
        
                Declare the End Of File identifer

        ***********************************************************************/

        enum {Eof = -1};

        /***********************************************************************
        
                Attach a filter to the IConduit input. The filter is invoked
                via its read(void[]) method whenever a block of content is 
                being read; the void[] array represents content destination.
                The filter should return the number of bytes it has actually
                produced: less than or equal to the length of the provided 
                array. 

                Filters are chained together such that the last filter added
                is the first one invoked. It is the responsibility of each
                filter to invoke the next link in the chain; for example:

                @code
                class MungingInputFilter : FilterInput
                {
                        int read (void[] dst)
                        {
                                // read the next X bytes
                                int count = next.read (dst);

                                // set everything to '*' !
                                dst[0..count] = '*';

                                // say how many we read
                                return count;
                        }
                }
                @endcode

                Notice how this filter invokes the 'next' instance before 
                munging the content ... the far end of the chain is where
                the original IConduit reader is attached, so it will get
                invoked eventually assuming each filter invokes 'next'. 
                If the next reader fails it will return IConduit.Eof, as
                should your filter (or throw an IOException). From a client 
                perspective, filters are attached like this:

                @code
                FileConduit fc = new FileConduit (...);

                fc.attach (new ZipInputFilter);
                fc.attach (new MungingInputFilter);
                @endcode

                Again, the last filter attached is the first one invoked 
                when a block of content is actually read. Each filter has
                two additional methods that it may use to control behavior:

                @code
                class FilterInput : IFilterInput
                {
                        protected IFilterInput next;

                        void bind (IFilterInput next)
                        {
                                this.next = next;
                        }

                        void unbind ()
                        {
                        }
                }
                @endcode

                The first method is invoked when the filter is attached to a 
                conduit, while the second is invoked just before the conduit 
                is closed. Both of these may be overridden by the filter for
                whatever purpose desired.

                Note that an input filter can choose to sidestep reading from 
                the conduit (per the usual case), and produce its input from 
                somewhere else entirely. This mechanism supports the notion
                of 'piping' between multiple conduits, or between a conduit 
                and something else entirely; it's a bridging mechanism.
                
        ***********************************************************************/

        void attach (IConduitSource input);

        /***********************************************************************
        
                Attach a filter to the IConduit output. The filter is invoked
                via its writ(void[]) method whenever a block of content is 
                being written; the void[] array supplies writeable content.
                The filter should return the number of bytes it has actually
                consumed: less than or equal to the length of the provided 
                array. 

                Filters are chained together such that the last filter added
                is the first one invoked. It is the responsibility of each
                filter to invoke the next link in the chain; for example:

                @code
                class MungingOutputFilter : FilterOutput
                {
                        int write (void[] src)
                        {
                                char[src.length] tmp;

                                // set everything to '*'
                                tmp = '*';

                                // write the munged output
                                return next.write (tmp);
                        }
                }
                @endcode

                Notice how the filter invokes the 'next' instance after 
                munging the content ... the far end of the chain is where
                the original IConduit writer is attached, so it will get
                invoked eventually assuming each filter invokes 'next'.
                If the next writer fails it will return IConduit.Eof, as
                should your filter (or throw an IOException). At the client 
                level, filters are attached like this:

                @code
                FileConduit fc = new FileConduit (...);

                fc.attach (new ZipOutputFilter);
                fc.attach (new MungingOutputFilter);
                @endcode

                Again, the last filter attached is the first one invoked 
                when a block of content is actually written. Each filter has
                two additional methods that it may use to control behavior:

                @code
                class FilterOutput : IFilterOutput
                {
                        protected IFilterOutput next;

                        void bind (IFilterOutput next)
                        {
                                this.next = next;
                        }

                        void unbind ()
                        {
                        }
                }
                @endcode

                The first method is invoked when the filter is attached to a 
                conduit, while the second is invoked just before the conduit 
                is closed. Both of these may be overridden by the filter for
                whatever purpose desired.

                Note that an output filter can choose to sidestep writing to 
                the conduit (per the usual case), and direct its output to 
                somewhere else entirely. This mechanism supports the notion
                of 'piping' between multiple conduits, or between a conduit 
                and something else entirely; as with the input-filter case,
                this is a bridging mechanism.

        ***********************************************************************/

        void attach (IConduitSink output);

        /***********************************************************************
                
                read from conduit into a target buffer

        ***********************************************************************/

        int read (IBuffer target);               

        /***********************************************************************
        
                write to conduit from a source buffer

        ***********************************************************************/

        int write (IBuffer source);               

        /***********************************************************************
                
                read from conduit into a target array

        ***********************************************************************/

        int read (void[] dst);               

        /***********************************************************************
        
                write to conduit from a source array

        ***********************************************************************/

        int write (void[] src);               

        /***********************************************************************
        
                Transfer the content of this conduit to another one. 
                Returns true if all content was successfully copied.
        
        ***********************************************************************/

        IConduit copy (IConduit source);

        /***********************************************************************
        
                Create a Buffer of a conduit-specific size

        ***********************************************************************/

        IBuffer createBuffer (); 
                     
        /***********************************************************************
        
                Returns true is this conduit can be read from

        ***********************************************************************/

        bool isReadable ();

        /***********************************************************************
        
                Returns true if this conduit can be written to

        ***********************************************************************/

        bool isWritable ();

        /***********************************************************************
        
                Returns true if this conduit is seekable (whether it 
                implements ISeekable)

        ***********************************************************************/

        bool isSeekable ();
}



/*******************************************************************************

        Defines an input-filter contract. The filter is invoked
        via its read(void[]) method whenever a block of content is 
        being read; the void[] array represents content destination.
        The filter should return the number of bytes it has actually
        produced: less than or equal to the length of the provided 
        array. 

        Filters are chained together such that the last filter added
        is the first one invoked. It is the responsibility of each
        filter to invoke the next link in the chain; for example:

        @code
        class MungingInputFilter : FilterInput
        {
                int read (void[] dst)
                {
                        // read the next X bytes
                        int count = next.read (dst);

                        // set everything to '*' !
                        dst[0..count] = '*';

                        // say how many we read
                        return count;
                }
        }
        @endcode

        Notice how this filter invokes the 'next' instance before 
        munging the content ... the far end of the chain is where
        the original IConduit reader is attached, so it will get
        invoked eventually assuming each filter invokes 'next'. 
        If the next reader fails it will return IConduit.Eof, as
        should your filter (or throw an IOException). From a client 
        perspective, filters are attached like this:

        @code
        FileConduit fc = new FileConduit (...);

        fc.attach (new ZipInputFilter);
        fc.attach (new MungingInputFilter);
        @endcode

        Again, the last filter attached is the first one invoked 
        when a block of content is actually read. Each filter has
        two additional methods that it may use to control behavior:

        @code
        class FilterInput : IFilterInput
        {
                protected IFilterInput next;

                void bind (IFilterInput next)
                {
                        this.next = next;
                }

                void unbind ()
                {
                }
        }
        @endcode

        The first method is invoked when the filter is attached to a 
        conduit, while the second is invoked just before the conduit 
        is closed. Both of these may be overridden by the filter for
        whatever purpose desired.

        Note that an input filter can choose to sidestep reading from 
        the conduit (per the usual case), and produce its input from 
        somewhere else entirely. This mechanism supports the notion
        of 'piping' between multiple conduits, or between a conduit 
        and something else entirely; it's a bridging mechanism.

*******************************************************************************/

interface IConduitSource
{
        /***********************************************************************
        
                conduit-specific reader

        ***********************************************************************/

        int read (void[] dst);               
                             
        /***********************************************************************
        
        ***********************************************************************/

        void bind (IConduitSource next);                       
                              
        /***********************************************************************
        
        ***********************************************************************/

        void unbind ();
}


/*******************************************************************************

        Defines an output-filter contract. The filter is invoked
        via its write(void[]) method whenever a block of content is 
        being written; the void[] array supplies writeable content.
        The filter should return the number of bytes it has actually
        consumed: less than or equal to the length of the provided 
        array. 

        Filters are chained together such that the last filter added
        is the first one invoked. It is the responsibility of each
        filter to invoke the next link in the chain; for example:

        @code
        class MungingOutputFilter : FilterOutput
        {
                int write (void[] src)
                {
                        char[src.length] tmp;

                        // set everything to '*'
                        tmp = '*';

                        // write the munged output
                        return next.write (tmp);
                }
        }
        @endcode

        Notice how the filter invokes the 'next' instance after 
        munging the content ... the far end of the chain is where
        the original IConduit writer is attached, so it will get
        invoked eventually assuming each filter invokes 'next'.
        If the next writer fails it will return IConduit.Eof, as
        should your filter (or throw an IOException). At the client 
        level, filters are attached like this:

        @code
        FileConduit fc = new FileConduit (...);

        fc.attach (new ZipOutputFilter);
        fc.attach (new MungingOutputFilter);
        @endcode

        Again, the last filter attached is the first one invoked 
        when a block of content is actually written. Each filter has
        two additional methods that it may use to control behavior:

        @code
        class FilterOutput : IFilterOutput
        {
                protected IFilterOutput next;

                void bind (IFilterOutput next)
                {
                        this.next = next;
                }

                void unbind ()
                {
                }
        }
        @endcode

        The first method is invoked when the filter is attached to a 
        conduit, while the second is invoked just before the conduit 
        is closed. Both of these may be overridden by the filter for
        whatever purpose desired.

        Note that an output filter can choose to sidestep writing to 
        the conduit (per the usual case), and direct its output to 
        somewhere else entirely. This mechanism supports the notion
        of 'piping' between multiple conduits, or between a conduit 
        and something else entirely; as with the input-filter case,
        this is a bridging mechanism.

*******************************************************************************/

interface IConduitSink
{
        /***********************************************************************
        
                conduit-specific writer

        ***********************************************************************/

        int write (void[] src);  
                             
        /***********************************************************************
        
        ***********************************************************************/

        void bind (IConduitSink next);                       
                              
        /***********************************************************************
        
        ***********************************************************************/

        void unbind ();
}




/*******************************************************************************

        Models the ability to seek within a conduit. 

*******************************************************************************/

interface ISeekable
{
        /***********************************************************************
        
                The anchor positions supported by ISeekable

        ***********************************************************************/

        enum SeekAnchor {
                        Begin   = 0,
                        Current = 1,
                        End     = 2,
                        };

        /***********************************************************************
                        
                Return the current conduit position (such as file position)
                
        ***********************************************************************/

        long getPosition ();

        /***********************************************************************
        
                Move the file position to the given offset (from the conduit 
                start) and return the adjusted position.

        ***********************************************************************/

        long seek (long offset);

        /***********************************************************************
        
                Move the file position to the given offset from the provided 
                anchor point, and return the adjusted position.

        ***********************************************************************/

        long seek (long offset, SeekAnchor anchor);
}

