/*******************************************************************************

        @file IWriter.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version; March 2004      

        @author         Kris
        @author         Ivan Senji (the "alias put" idea)


*******************************************************************************/

module mango.io.model.IWriter;

public import mango.io.model.IBuffer;


/*******************************************************************************

*******************************************************************************/

typedef void delegate (void* dst, uint bytes, int type) BufferEncoder;

/*******************************************************************************

        Interface to make any class compatible with any IWriter.

*******************************************************************************/

interface IWritable
{
        void write (IWriter input);
}

/*******************************************************************************

        Any class implementing IEncoder can be bound to a writer using
        the bind() method.
        
*******************************************************************************/

interface IEncoder
{
        ConverterType type ();

        BufferEncoder bind (IWriter output);
}

/*******************************************************************************

        Make a signature for IWritable classes to use when they wish to 
        avoid being processed by decorating writers, such as TextWriter.

*******************************************************************************/

interface IPhantomWriter : IWritable
{
}

/*******************************************************************************

        Make a signature for Newline classes

*******************************************************************************/

interface INewlineWriter : IPhantomWriter {}

/*******************************************************************************

        IWriter interface. IWriter provides the means to append formatted 
        data to an IBuffer, and exposes a convenient method of handling a
        variety of data types. In addition to writing native types such
        as integer and char[], writers also process any class which has
        implemented the IWritable interface (one method).

*******************************************************************************/

interface IWriter
{
        // alias the put() methods for IOStream and Whisper styles
        alias put       opShl;
        alias put       opCall;

        /***********************************************************************
        
                These are the basic writer methods

        ***********************************************************************/

        IWriter put (bool x);
        IWriter put (ubyte x);
        IWriter put (byte x);
        IWriter put (ushort x);
        IWriter put (short x);
        IWriter put (uint x);
        IWriter put (int x);
        IWriter put (ulong x);
        IWriter put (long x);
        IWriter put (float x);
        IWriter put (double x);
        IWriter put (real x);
        IWriter put (char x);
        IWriter put (wchar x);
        IWriter put (dchar x);

        IWriter put (byte[] x);
        IWriter put (short[] x);
        IWriter put (int[] x);
        IWriter put (long[] x);
        IWriter put (ubyte[] x);
        IWriter put (ushort[] x);
        IWriter put (uint[] x);
        IWriter put (ulong[] x);
        IWriter put (float[] x);
        IWriter put (double[] x);
        IWriter put (real[] x);
        IWriter put (char[] x);

        // these two will have to wait until Walter fixes 
        // the string-literal conversion fiasco ...
        IWriter putw (wchar[] x);
        IWriter putd (dchar[] x);

        /***********************************************************************
        
                This is the mechanism used for binding arbitrary classes 
                to the IO system. If a class implements IWritable, it can
                be used as a target for IWriter put() operations. That is, 
                implementing IWritable is intended to transform any class 
                into an IWriter adaptor for the content held therein.

        ***********************************************************************/

        IWriter put (IWritable x);

        /***********************************************************************
        
                Bind an IEncoder to the writer. Encoders are intended to
                be used as a conversion mechanism between various character
                representations (encodings). Each type may be configured with 
                a distinct encoder.

                An appropriate encoder set should be attached to each 
                IWriter, and thus be available for subsequent use. A raw 
                binary implementation is attached by default (no encoding).

                See module mango.icu.UMango for an example of encoder 
                implementation -- those classes bind the ICU converters 
                to this IO package.

        ***********************************************************************/

        void setEncoder (IEncoder e); 

        /***********************************************************************

                Output a newline. Do this indirectly so that it can be 
                intercepted by subclasses.
        
        ***********************************************************************/

        IWriter cr ();

        /***********************************************************************
        
                Flush the output of this writer. Throws an IOException 
                if the operation fails. These are aliases for each other.

        ***********************************************************************/

        IWriter put ();
        IWriter flush ();

        /***********************************************************************
        
                Return the associated buffer

        ***********************************************************************/

        IBuffer getBuffer ();
}

